#
# GtkHelp.py -- customized Gtk4 widgets
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import os.path
import math
import random
import time

import numpy as np

from ginga.misc import Bunch, Callback
from ginga.fonts import font_asst
from ginga.util.paths import icondir, app_icon_path
import ginga.toolkit

import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk  # noqa
from gi.repository import Gdk  # noqa
from gi.repository import GdkPixbuf  # noqa
from gi.repository import GObject  # noqa
from gi.repository import Pango  # noqa
import cairo

ginga.toolkit.use('gtk4')

DND_TARGET_TYPE_TEXT = 0
DND_TARGET_TYPE_URIS = 1


class WidgetMask(object):
    def __init__(self, *args):
        self.cb_fn = None
        self.cb_args = []
        self.cb_kwdargs = {}

        self.connected = False
        self.changed = False

    def sconnect(self, signal, cb_fn, *args, **kwdargs):
        self.cb_fn = cb_fn
        self.cb_args = args
        self.cb_kwdargs = kwdargs

        self.connect(signal, self.cb)
        self.connected = True

    def change(self):
        if self.connected:
            self.changed = True

    def cb(self, *args):
        if self.changed:
            self.changed = False
            return

        newargs = list(args)
        newargs.extend(self.cb_args)
        kwdargs = self.cb_kwdargs.copy()
        return self.cb_fn(*newargs, **kwdargs)


class TopLevel(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self)


class CheckButton(WidgetMask, Gtk.CheckButton):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.CheckButton.__init__(self, *args, **kwdargs)

    def set_active(self, newval):
        oldval = self.get_active()
        if oldval != newval:
            self.change()

        super(CheckButton, self).set_active(newval)


class ToggleButton(WidgetMask, Gtk.ToggleButton):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.ToggleButton.__init__(self, *args, **kwdargs)

    def set_active(self, newval):
        oldval = self.get_active()
        if oldval != newval:
            self.change()

        super(ToggleButton, self).set_active(newval)

    def toggle(self):
        oldval = self.get_active()
        newval = not oldval
        super(ToggleButton, self).set_active(newval)


class RadioButton(WidgetMask, Gtk.CheckButton):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.CheckButton.__init__(self, *args, **kwdargs)

    def set_active(self, newval):
        oldval = self.get_active()
        if oldval != newval:
            self.change()

        super(RadioButton, self).set_active(newval)

    def toggle(self):
        oldval = self.get_active()
        newval = not oldval
        super(RadioButton, self).set_active(newval)


# class CheckMenuItem(WidgetMask, Gtk.CheckMenuItem):
#     def __init__(self, *args, **kwdargs):
#         WidgetMask.__init__(self)
#         Gtk.CheckMenuItem.__init__(self, *args, **kwdargs)

#     def set_active(self, newval):
#         oldval = self.get_active()
#         if oldval != newval:
#             self.change()

#         super(CheckMenuItem, self).set_active(newval)


class SpinButton(WidgetMask, Gtk.SpinButton):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.SpinButton.__init__(self, *args, **kwdargs)

    def set_value(self, newval):
        oldval = self.get_value()
        if oldval != newval:
            self.change()

        super(SpinButton, self).set_value(newval)


class HScale(WidgetMask, Gtk.Scale):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.Scale.__init__(self, Gtk.ORIENTATION_HORIZONTAL, *args, **kwdargs)

    def set_value(self, newval):
        oldval = self.get_value()
        if oldval != newval:
            self.change()

        super(HScale, self).set_value(newval)


class VScale(WidgetMask, Gtk.Scale):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.Scale.__init__(self, Gtk.ORIENTATION_VERTICAL, *args, **kwdargs)

    def set_value(self, newval):
        oldval = self.get_value()
        if oldval != newval:
            self.change()

        super(VScale, self).set_value(newval)


class ComboBox(WidgetMask, Gtk.ComboBox):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.ComboBox.__init__(self, *args, **kwdargs)

    def set_active(self, newval):
        oldval = self.get_active()
        if oldval != newval:
            self.change()

        super(ComboBox, self).set_active(newval)

    def insert_alpha(self, text):
        model = self.get_model()
        tup = (text, )
        j = 0
        for i in range(len(model)):
            j = i
            if model[i][0] > text:
                model.insert(j, tup)
                return
        model.insert(j + 1, tup)

    def insert_text(self, idx, text):
        model = self.get_model()
        tup = (text, )
        model.insert(idx, tup)

    def delete_alpha(self, text):
        model = self.get_model()
        for i in range(len(model)):
            if model[i][0] == text:
                del model[i]
                return

    def clear(self):
        model = self.get_model()
        model.clear()

    def show_text(self, text):
        model = self.get_model()
        for i in range(len(model)):
            if model[i][0] == text:
                self.set_active(i)
                return


class Notebook(WidgetMask, Gtk.Notebook):
    def __init__(self, *args, **kwdargs):
        WidgetMask.__init__(self)
        Gtk.Notebook.__init__(self, *args, **kwdargs)

    def set_group_id(self, id):
        super(Notebook, self).set_group_name(str(id))

    def set_current_page(self, new_idx):
        old_idx = self.get_current_page()
        if old_idx != new_idx:
            self.change()

        super(Notebook, self).set_current_page(new_idx)


class MultiDragDropTreeView(Gtk.TreeView):
    '''TreeView that captures mouse events to make drag and drop work
    properly
    See: https://gist.github.com/kevinmehall/278480#file-multiple-selection-dnd-class-py
    '''

    def __init__(self):
        super(MultiDragDropTreeView, self).__init__()

        event = Gtk.GestureClick.new()
        event.connect("pressed", self.on_button_press)
        # all buttons
        event.set_button(0)
        self.add_controller(event)
        event = Gtk.GestureClick.new()
        event.connect("released", self.on_button_release)
        event.set_button(0)
        self.add_controller(event)
        self.defer_select = False

    def on_button_press(self, widget, event):
        # Here we intercept mouse clicks on selected items so that we can
        # drag multiple items without the click selecting only one
        target = self.get_path_at_pos(int(event.x), int(event.y))
        if (target and
                event.type == Gdk.EventType.BUTTON_PRESS and
                not (event.state &
                     (Gdk.ModifierType.CONTROL_MASK |
                      Gdk.ModifierType.SHIFT_MASK)) and
                self.get_selection().path_is_selected(target[0])):
            # disable selection
            self.get_selection().set_select_function(lambda *ignore: False)
            self.defer_select = target[0]

    def on_button_release(self, widget, event):
        # re-enable selection
        self.get_selection().set_select_function(lambda *ignore: True)

        target = self.get_path_at_pos(int(event.x), int(event.y))
        if (self.defer_select and target and
                self.defer_select == target[0] and
                not (event.x == 0 and event.y == 0)):  # certain drag and drop
            self.set_cursor(target[0], target[1], False)

        self.defer_select = False


class MDISubWindow(Callback.Callbacks):

    def __init__(self, widget, label, iconpath=None):
        super(MDISubWindow, self).__init__()

        self.widget = widget
        self.border_wd = 6

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        set_border_width(vbox, self.border_wd)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        hbox.set_spacing(2)

        # set window icon
        if iconpath is None:
            iconpath = app_icon_path
        pixbuf = pixbuf_new_from_file_at_size(iconpath, 32, 32)
        self.image = Gtk.Image.new_from_pixbuf(pixbuf)
        hbox.append(self.image)

        modify_bg(label, "gray90")
        self.title = label
        label.set_hexpand(True)
        hbox.append(label)

        # titlebar buttons
        iconfile = os.path.join(icondir, "close.svg")
        pixbuf = pixbuf_new_from_file_at_size(iconfile, 24, 24)
        image = Gtk.Image.new_from_pixbuf(pixbuf)
        close = Gtk.Button()
        modify_bg(close, "gray90")
        close.set_child(image)
        iconfile = os.path.join(icondir, "maximize.svg")
        pixbuf = pixbuf_new_from_file_at_size(iconfile, 24, 24)
        image = Gtk.Image.new_from_pixbuf(pixbuf)
        maxim = Gtk.Button()
        modify_bg(maxim, "gray90")
        maxim.set_child(image)
        iconfile = os.path.join(icondir, "minimize.svg")
        pixbuf = pixbuf_new_from_file_at_size(iconfile, 24, 24)
        image = Gtk.Image.new_from_pixbuf(pixbuf)
        minim = Gtk.Button()
        modify_bg(minim, "gray90")
        minim.set_child(image)
        hbox.append(minim)
        hbox.append(maxim)
        hbox.append(close)

        vbox.append(hbox)
        widget.set_vexpand(True)
        widget.set_hexpand(True)

        # what size does the widget want to be?
        wd, ht = widget.get_width(), widget.get_height()
        wd += self.border_wd * 2
        ht += self.border_wd * 2
        self.x, self.y = 0, 0
        self.width, self.height = max(wd, 300), max(ht, 300)

        vbox.set_size_request(self.width, self.height)
        vbox.append(widget)

        frame = Gtk.Frame()
        modify_bg(frame, "gray70")
        frame.set_child(vbox)
        self.frame = frame

        for name in ('close', 'maximize', 'minimize'):
            self.enable_callback(name)

        maxim.connect('clicked', lambda *args: self.make_callback('maximize'))
        minim.connect('clicked', lambda *args: self.make_callback('minimize'))
        close.connect('clicked', lambda *args: self.make_callback('close'))

    def set_icon(self, iconpath):
        pixbuf = pixbuf_new_from_file_at_size(iconpath, 32, 32)
        self.image.set_from_pixbuf(pixbuf)

    def raise_(self):
        # TODO!
        #self.frame.raise_()
        parent = self.frame.get_parent()
        parent.raise_subwin(self)

    def lower(self):
        # TODO
        #self.frame.lower()
        parent = self.frame.get_parent()
        parent.lower_subwin(self)

    def focus(self):
        self.frame.grab_focus()


class MDIWidget(Gtk.Fixed):
    """
    Multiple Document Interface type widget for Gtk.
    """
    def __init__(self):
        Gtk.Fixed.__init__(self)

        self.children = []
        self.cur_index = -1
        self.selected_child = None
        self.kbdmouse_mask = 0
        self.cascade_offset = 50
        self.minimized_width = 150
        self.delta_px = 50

        controller = Gtk.EventControllerMotion.new()
        controller.connect("motion", self.motion_notify_event)
        self.add_controller(controller)
        controller = Gtk.GestureClick.new()
        controller.connect("pressed", self.button_press_event)
        controller.connect("released", self.button_release_event)
        # all buttons
        controller.set_button(0)
        self.add_controller(controller)
        self._button_controller = controller
        event = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.BOTH_AXES)
        event.connect("scroll", self.scroll_event)
        self.add_controller(event)

        modify_bg(self, "gray50")

    def add_subwin(self, subwin):

        self.children.append(subwin)

        event = Gtk.GestureClick.new()
        event.connect("pressed", self.select_child_cb, subwin)
        subwin.title.add_controller(event)
        event = Gtk.GestureClick.new()
        event.connect("pressed", self.start_resize_cb, subwin)
        subwin.frame.add_controller(event)

        subwin.add_callback('maximize', lambda *args: self.maximize_page(subwin))
        subwin.add_callback('minimize', lambda *args: self.minimize_page(subwin))

        self.put(subwin.frame, subwin.x, subwin.y)

        # note: seem to need a slight delay to let the widget be mapped
        # in order to accurately determine its position and size
        #self.update_subwin_position(subwin)
        #self.update_subwin_size(subwin)
        GObject.timeout_add(1000, self.update_subwin_position, subwin)
        GObject.timeout_add(1500, self.update_subwin_size, subwin)

        self._update_area_size()

    def append_page(self, widget, label):

        subwin = MDISubWindow(widget, label)

        # pick a random spot to place the window initially
        wd, ht = self.get_width(), self.get_height()
        x = random.randint(self.cascade_offset,  # nosec
                           max(self.cascade_offset + 10, wd // 2))
        y = random.randint(self.cascade_offset,  # nosec
                           max(self.cascade_offset + 10, ht // 2))
        subwin.x, subwin.y = x, y

        self.add_subwin(subwin)
        return subwin

    def set_tab_reorderable(self, w, tf):
        pass

    def set_tab_detachable(self, w, tf):
        pass

    def get_tab_label(self, w):
        return None

    def page_num(self, widget):
        index, subwin = self._widget_to_index(widget)
        return index

    def get_nth_page(self, idx):
        if 0 <= idx < len(self.children):
            subwin = self.children[idx]
            return subwin.widget
        return None

    def set_current_page(self, idx):
        subwin = self.children[idx]
        subwin.raise_()
        self.cur_index = idx

    def get_current_page(self):
        return self.cur_index

    def _widget_to_index(self, widget):
        index = 0
        for subwin in self.children:
            if subwin.widget == widget:
                return index, subwin
            index += 1
        return -1, None

    def remove_page(self, idx):
        subwin = self.children[idx]
        self.remove(subwin.widget)

    def remove(self, widget):
        idx, subwin = self._widget_to_index(widget)
        if subwin is not None:
            self.children.remove(subwin)
            self.cur_index = -1
            frame = subwin.frame
            #super(MDIWidget, self).remove(frame)
            widget.unparent()
        self._update_area_size()

    def get_widget_position(self, widget):
        x, y = self.get_child_position(widget)
        x, y = int(x), int(y)
        return x, y

    def get_widget_size(self, widget):
        width, height = widget.get_width(), widget.get_height()
        return width, height

    def update_subwin_position(self, subwin):
        x, y = self.get_widget_position(subwin.frame)
        subwin.x, subwin.y = x, y

    def update_subwin_size(self, subwin):
        wd, ht = self.get_widget_size(subwin.frame)
        subwin.width, subwin.height = wd, ht

    def raise_subwin(self, subwin):
        self.update_subwin_position(subwin)
        super().remove(subwin.frame)
        self.put(subwin.frame, subwin.x, subwin.y)

    def lower_subwin(self, subwin):
        #subwin.raise_()
        pass

    def select_child_cb(self, gesture, whatsit, x, y, subwin):
        x_root, y_root = subwin.title.translate_coordinates(self, x, y)

        x, y = self.get_widget_position(subwin.frame)
        subwin.x, subwin.y = x, y

        # make this the selected widget
        idx = self.page_num(subwin.widget)
        if idx >= 0:
            self.set_current_page(idx)

        self.selected_child = Bunch.Bunch(subwin=subwin, action='move',
                                          x_origin=x, y_origin=y,
                                          x_root=x_root, y_root=y_root)
        #gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        return True

    def start_resize_cb(self, gesture, whatit, x, y, subwin):
        if self.selected_child is not None:
            # move is being done
            return True
        self.update_subwin_size(subwin)

        x_root, y_root = subwin.frame.translate_coordinates(self, x, y)
        x, y = x_root, y_root

        x1, y1 = self.get_widget_position(subwin.frame)
        wd, ht = self.get_widget_size(subwin.frame)
        x2, y2 = x1 + wd, y1 + ht
        subwin.x, subwin.y = x1, y1
        subwin.width, subwin.height = wd, ht

        updates = set([])
        if abs(x - x2) < self.delta_px:
            # right side
            if abs(y - y2) < self.delta_px:
                # lower right corner
                origin = 'lr'
                updates = set(['w', 'h'])
            elif abs(y - y1) < self.delta_px:
                origin = 'ur'
                updates = set(['w', 'h', 'y'])
            else:
                origin = 'r'
                updates = set(['w'])
        elif abs(x - x1) < self.delta_px:
            # left side
            if abs(y - y2) < self.delta_px:
                # lower left corner
                origin = 'll'
                updates = set(['w', 'h', 'x'])
            elif abs(y - y1) < self.delta_px:
                origin = 'ul'
                updates = set(['w', 'h', 'x', 'y'])
            else:
                origin = 'l'
                updates = set(['w', 'x'])
        elif abs(y - y2) < self.delta_px:
            # bottom
            origin = 'b'
            updates = set(['h'])
        else:
            origin = 't'
            updates = set(['h', 'y'])

        self.selected_child = Bunch.Bunch(subwin=subwin, action='resize',
                                          x_origin=x1, y_origin=y1,
                                          wd=wd, ht=ht,
                                          x_root=x_root, y_root=y_root,
                                          origin=origin, updates=updates)
        #gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        return True

    def button_press_event(self, gesture, whatsit, x, y):
        # event_button = gesture.get_current_button()
        # button = 0
        # if event_button != 0:
        #     button |= 0x1 << (event_button - 1)
        #self._button_controller.set_state(Gtk.EventSequenceState.CLAIMED)
        return True

    def _update_area_size(self):
        mx_wd, mx_ht = self.get_width(), self.get_height()

        for subwin in self.children:
            x, y = self.get_widget_position(subwin.frame)
            wd, ht = self.get_widget_size(subwin.frame)

            mx_wd, mx_ht = max(mx_wd, x + wd), max(mx_ht, y + ht)

        self.set_size(mx_wd, mx_ht)

    def _resize(self, bnch, x_root, y_root):
        subwin = bnch.subwin
        updates = bnch.updates

        dx, dy = x_root - bnch.x_root, y_root - bnch.y_root

        wd = bnch.wd
        if 'w' in updates:
            wd = int(wd + dx)
        ht = bnch.ht
        if 'h' in updates:
            ht = int(ht + dy)

        if 'x' in updates or 'y' in updates:
            x = bnch.x_origin
            if 'x' in updates:
                x = int(x + dx)
                if x < bnch.x_origin:
                    wd = bnch.wd + abs(dx)
                else:
                    wd = bnch.wd + -abs(dx)

            y = bnch.y_origin
            if 'y' in updates:
                y = int(y + dy)
                if y < bnch.y_origin:
                    ht = bnch.ht + abs(dy)
                else:
                    ht = bnch.ht + -abs(dy)

            # this works better if it is not self.move_page()
            self.move(subwin.frame, x, y)

        if 'w' in updates or 'h' in updates:
            # this works better if it is not self.resize_page()
            subwin.frame.set_size_request(wd, ht)

        self._update_area_size()

    def button_release_event(self, gesture, whatsit, x, y):
        event_button = gesture.get_current_button()
        x_root, y_root = x, y
        button = 0
        if event_button != 0:
            button |= 0x1 << (event_button - 1)

        if self.selected_child is not None:
            bnch = self.selected_child
            subwin = bnch.subwin
            if bnch.action == 'move':
                x = int(subwin.x + (x_root - bnch.x_root))
                y = int(subwin.y + (y_root - bnch.y_root))
                self.move_page(subwin, x, y)

            elif bnch.action == 'resize':
                self._resize(bnch, x_root, y_root)

                self.update_subwin_position(subwin)
                # NOTE: necessary for wrapped widget to remember position
                self.move_page(subwin, subwin.x, subwin.y)

                self.update_subwin_size(subwin)
                # NOTE: necessary for wrapped widget to remember size
                self.resize_page(subwin, subwin.width, subwin.height)

            self.selected_child = None

        self._update_area_size()
        self._button_controller.set_state(Gtk.EventSequenceState.CLAIMED)
        return True

    def motion_notify_event(self, motion, x, y):
        button = 0
        x_root, y_root = x, y
        state = motion.get_current_event_state()

        if state & Gdk.ModifierType.BUTTON1_MASK:
            button |= 0x1
        elif state & Gdk.ModifierType.BUTTON2_MASK:
            button |= 0x2
        elif state & Gdk.ModifierType.BUTTON3_MASK:
            button |= 0x4

        if (button & 0x1) and (self.selected_child is not None):
            bnch = self.selected_child
            subwin = bnch.subwin
            if bnch.action == 'move':
                x = int(subwin.x + (x_root - bnch.x_root))
                y = int(subwin.y + (y_root - bnch.y_root))
                # this works better if it is not self.move_page()
                self.move(subwin.frame, x, y)

            elif bnch.action == 'resize':
                self._resize(bnch, x_root, y_root)

        self._update_area_size()
        #self._motion_controller.set_state(Gtk.EventSequenceState.CLAIMED)
        return True

    def scroll_event(self, controller, dx, dy):
        # disable wheel event scrolling the space
        return True

    def tile_pages(self):
        # calculate number of rows and cols, try to maintain a square
        # TODO: take into account the window geometry
        num_widgets = len(self.children)
        rows = int(round(math.sqrt(num_widgets)))
        cols = rows
        if rows**2 < num_widgets:
            cols += 1

        # find out how big each window should be
        width, height = self.get_width(), self.get_height()
        wd, ht = width // cols, height // rows

        # and move and resize them into place
        for i in range(0, rows):
            for j in range(0, cols):
                index = i * cols + j
                if index < num_widgets:
                    subwin = self.children[index]

                    self.resize_page(subwin, wd, ht)

                    x, y = j * wd, i * ht
                    self.move_page(subwin, x, y)

                    subwin.raise_()

        self._update_area_size()

    def cascade_pages(self):
        x, y = 0, 0
        for subwin in self.children:
            self.move_page(subwin, x, y)
            subwin.raise_()
            x += self.cascade_offset
            y += self.cascade_offset

        self._update_area_size()

    def use_tabs(self, tf):
        pass

    def set_size(self, wd, ht):
        self.set_size_request(wd, ht)

    def move_page(self, subwin, x, y):
        self.move(subwin.frame, x, y)
        subwin.x, subwin.y = x, y

    def resize_page(self, subwin, wd, ht):
        subwin.frame.set_size_request(wd, ht)
        subwin.width, subwin.height = wd, ht

    def maximize_page(self, subwin):
        widget = self.get_parent()
        wd, ht = widget.get_width(), widget.get_height()

        subwin.raise_()
        self.resize_page(subwin, wd, ht)
        self.move_page(subwin, 0, 0)

        self._update_area_size()

    def minimize_page(self, subwin):
        widget = self.get_parent()
        height = widget.get_height()

        x, _ = self.get_widget_position(subwin.frame)

        _, ht = self.get_widget_size(subwin.title)

        self.resize_page(subwin, self.minimized_width, ht)
        self.move_page(subwin, x, height - ht)
        subwin.lower()

        self._update_area_size()

    def close_page(self, subwin):
        self._update_area_size()


class Splitter(Gtk.Fixed):
    """
    Splitter type widget for Gtk.
    """
    def __init__(self, orientation='horizontal', thumb_px=8):
        Gtk.Fixed.__init__(self)

        self.orientation = orientation
        self._sizes = []
        self._dims = (0, 0)
        self.children = []
        self.thumbs = []
        self.thumb_px = thumb_px
        self.thumb_aspect = 3.25
        self.kbdmouse_mask = 0

        self.set_hexpand(True)
        self.set_vexpand(True)
        #self.connect("size-allocate", self._size_allocate_cb)
        self.connect("notify", self._size_allocate_cb)

    def add_widget(self, widget):
        rect = self.get_allocation()
        wd, ht = rect.width, rect.height

        self.children.append(widget)

        if len(self.children) == 1:
            widget.set_size_request(wd, ht)
            self.put(widget, 0, 0)
            sizes = self._sizes
            if len(sizes) == 0:
                pos = wd if self.orientation == 'horizontal' else ht
                sizes = [pos]
            self.set_sizes(sizes)

        else:
            if self.orientation == 'horizontal':
                thumbfile, _w, _h = ('vdots.svg', self.thumb_px,
                                     int(self.thumb_px * self.thumb_aspect))
                widget.set_hexpand(True)
            else:
                thumbfile, _w, _h = ('hdots.svg',
                                     int(self.thumb_px * self.thumb_aspect),
                                     self.thumb_px)
                widget.set_vexpand(True)
            iconfile = os.path.join(icondir, thumbfile)
            pixbuf = pixbuf_new_from_file_at_size(iconfile, _w, _h)
            image = Gtk.Image.new_from_pixbuf(pixbuf)
            #modify_bg(thumb, "gray90")
            thumb = image

            i = len(self.thumbs)
            self.thumbs.append(thumb)
            event = Gtk.EventControllerMotion.new()
            event.connect("enter", self._thumb_enter_cb)
            thumb.add_controller(event)
            event = Gtk.EventControllerMotion.new()
            event.connect("leave", self._thumb_leave_cb)
            thumb.add_controller(event)
            event = Gtk.EventControllerMotion.new()
            event.connect("motion", self._do_resize_cb, thumb, i)
            thumb.add_controller(event)
            event = Gtk.GestureClick.new()
            event.connect("pressed", self._start_resize_cb, thumb, i)
            # all buttons
            event.set_button(0)
            thumb.add_controller(event)
            event = Gtk.GestureClick.new()
            event.connect("released", self._stop_resize_cb, thumb, i)
            event.set_button(0)
            thumb.add_controller(event)

            self.put(thumb, 0, 0)
            self.put(widget, 0, 0)

            sizes = self._sizes
            if len(sizes) < len(self.children):
                pos = wd if self.orientation == 'horizontal' else ht
                sizes.append(pos)
            self.set_sizes(sizes)

    def _thumb_enter_cb(self, controller, x, y):
        # change the cursor to a resize one when we enter the thumb area
        display = self.get_display()
        cur_name = ('ew-resize' if self.orientation == 'horizontal'
                    else 'ns-resize')
        cursor = Gdk.Cursor.new_from_name(cur_name)
        self.set_cursor(cursor)

    def _thumb_leave_cb(self, controller):
        # change the cursor to the normal one when we leave the thumb area
        cursor = Gdk.Cursor.new_from_name('default')
        self.set_cursor(cursor)

    def get_sizes(self):
        return list(self._sizes)

    def set_sizes(self, sizes):
        sizes = list(sizes)
        ## if sizes == self._sizes:
        ##     return
        if self.get_realized():
            rect = self.get_allocation()
            wd, ht = rect.width, rect.height
        else:
            min_req, nat_req = self.get_preferred_size()
            wd, ht = nat_req.width, nat_req.height

        x, y = 0, 0
        # calc space needed by all necessary thumbs
        remaining_thumb_space = max(0, len(self.children) - 1) * self.thumb_px
        new_sizes = []
        thumbs, widgets = [], []
        for num, child in enumerate(self.children):
            off = sizes[num]

            if self.orientation == 'horizontal':
                if num == 0:
                    widgets.append((child, 0, 0, off, ht))
                    new_sizes.append(off)
                    x += off

                else:
                    thumb = self.thumbs[num - 1]
                    thumbs.append((thumb, x, y, self.thumb_px, ht))
                    x += self.thumb_px
                    remaining_thumb_space -= self.thumb_px

                    rest = max(0, wd - (x + remaining_thumb_space))
                    if num < len(self.children) - 1:
                        rest = min(off, rest)
                    widgets.append((child, x, y, rest, ht))
                    new_sizes.append(rest)
                    x += rest

            else:
                if num == 0:
                    widgets.append((child, 0, 0, wd, off))
                    new_sizes.append(off)
                    y += off

                else:
                    thumb = self.thumbs[num - 1]
                    thumbs.append((thumb, x, y, wd, self.thumb_px))
                    y += self.thumb_px
                    remaining_thumb_space -= self.thumb_px

                    rest = max(0, ht - (y + remaining_thumb_space))
                    if num < len(self.children) - 1:
                        rest = min(off, rest)
                    widgets.append((child, x, y, wd, rest))
                    new_sizes.append(rest)
                    y += rest

        self._sizes = new_sizes
        assert len(self._sizes) == len(self.children)

        for child, x, y, wd, ht in widgets:
            self._move_resize_child(child, x, y, wd, ht)
        for thumb, x, y, wd, ht in thumbs:
            self._move_resize_child(thumb, x, y, wd, ht)

    def remove(self, child):
        if child not in self.children:
            raise ValueError("widget is not one of our children")
        idx = self.children.index(child)
        if len(self.children) > 1:
            if idx > 0:
                # not first child
                thumb = self.thumbs.pop(idx - 1)
            else:
                thumb = self.thumbs.pop(0)

            super(Splitter, self).remove(thumb)

        self._sizes.pop(idx)
        self.children.remove(child)
        super(Splitter, self).remove(child)

        self.set_sizes(self._sizes)

    def _move_resize_child(self, child, x, y, wd, ht):
        rect = child.get_allocation()
        modified = False

        if (rect.x, rect.y) != (x, y):
            modified = True
            self.move(child, x, y)

        if (rect.width, rect.height) != (wd, ht):
            modified = True
            child.set_size_request(wd, ht)

            alloc = Gdk.Rectangle()
            alloc.x, alloc.y, alloc.width, alloc.height = x, y, wd, ht
            child.size_allocate(alloc, -1)

            #child.set_clip(alloc)

            # win = child.get_window()
            # if win is not None:
            #     win.invalidate_rect(None, True)
            #     win.resize(wd, ht)
            #child.resize(wd, ht)

        if modified:
            # don't think this should be necessary, but just in case
            child.queue_draw()
            child.queue_resize()
            child.queue_allocate()

    def _calc_size(self, i, pos):
        sizes = list(self._sizes)
        n = sum([sizes[j] for j in range(0, i)])
        n += max(0, i - 1) * self.thumb_px
        return max(0, pos - n)

    def _start_resize_cb(self, gesture, whatsit, x, y, widget, i):
        x, y = widget.translate_coordinates(self, x, y)

        pos = x if self.orientation == 'horizontal' else y
        sizes = list(self._sizes)
        sizes[i] = self._calc_size(i, pos)
        self.set_sizes(sizes)
        return True

    def _stop_resize_cb(self, gesture, whatsit, x, y, widget, i):
        x, y = widget.translate_coordinates(self, x, y)

        pos = x if self.orientation == 'horizontal' else y
        sizes = list(self._sizes)
        sizes[i] = self._calc_size(i, pos)
        self.set_sizes(sizes)
        return True

    def _do_resize_cb(self, motion, x, y, widget, i):
        button = self.kbdmouse_mask
        state = motion.get_current_event_state()
        x, y = widget.translate_coordinates(self, x, y)

        if state & Gdk.ModifierType.BUTTON1_MASK:
            button |= 0x1
        elif state & Gdk.ModifierType.BUTTON2_MASK:
            button |= 0x2
        elif state & Gdk.ModifierType.BUTTON3_MASK:
            button |= 0x4

        if button == 0x1:
            pos = x if self.orientation == 'horizontal' else y
            sizes = list(self._sizes)
            sizes[i] = self._calc_size(i, pos)
            self.set_sizes(sizes)
        return True

    #def _size_allocate_cb(self, widget, wd, ht):
    def _size_allocate_cb(self, widget, param):
        #x, y, wd, ht = rect.x, rect.y, rect.width, rect.height
        if param.name not in ['width-request', 'height-request']:
            return
        wd, ht = self.get_width(), self.get_height()
        dims = (wd, ht)
        if dims == self._dims:
            return
        #super(Splitter, self).set_size(wd, ht)

        self._dims = dims

        self.set_sizes(self._sizes)
        return True


class Dial(Gtk.DrawingArea):

    __gtype_name__ = "Dial"

    __gsignals__ = {
        "value-changed": (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE,
                          (GObject.TYPE_FLOAT,)),
    }

    def __init__(self):
        Gtk.DrawingArea.__init__(self)
        self.set_can_focus(True)
        self.dims = np.array((0, 0))
        self.center = np.array((0.0, 0.0))
        self.bg = (0.94, 0.94, 0.94)
        self.fg = (0.4, 0.4, 0.4)
        self.knob_fg = (1.0, 1.0, 1.0)
        self.knob_fill = (0.2, 0.2, 0.2)
        self.focus_fg = (0.2, 0.6, 0.9)
        self.fontname = 'Sans Serif'
        self.fontsize = 10.0
        self._has_focus = False
        self.surface = None

        # draw labels
        self.draw_scale = True
        # how to rotate the labels
        self.label_style = 1
        self.values = []
        self.draw_value_pos = 0

        self.value = 0.0
        self.value_text = str(self.value)

        # internal state
        self._dragging = False
        self.tracking = False
        self.wrap = False
        self.angle = 0.0
        self.ang_offset = 0.0
        self.ang_invert = False
        self.turn_delta = 6.0
        self.min_ang_deg = 0.0
        self.max_ang_deg = 360.0

        self.connect("draw", self.draw_event)
        self.connect("configure-event", self.configure_event)
        self.set_app_paintable(True)
        # prevents extra redraws, because we manually redraw on a size
        # change
        self.set_redraw_on_allocate(False)

        self.connect('button-press-event', self.button_press_event)
        self.connect('button-release-event', self.button_release_event)
        self.connect('motion-notify-event', self.motion_notify_event)
        self.connect('scroll-event', self.scroll_event)
        self.connect('focus_in_event', self.focus_event, True)
        self.connect('focus_out_event', self.focus_event, False)
        mask = self.get_events()
        self.set_events(mask |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK |
                        Gdk.EventMask.POINTER_MOTION_MASK |
                        Gdk.EventMask.SCROLL_MASK |
                        Gdk.EventMask.FOCUS_CHANGE_MASK |
                        Gdk.EventMask.EXPOSURE_MASK)

    def button_press_event(self, widget, event):
        if event.button == 1:
            self._dragging = True
            self._calc_action(event.x, event.y)
            return True

    def button_release_event(self, widget, event):
        self._dragging = False
        self._calc_action(event.x, event.y)
        return True

    def motion_notify_event(self, widget, event):
        # Are we holding down the left mouse button?
        if not self._dragging:
            return False

        self._calc_action(event.x, event.y)
        return True

    def scroll_event(self, widget, event):
        degrees, direction = get_scroll_info(event)
        if direction < 180.0:
            self.turn_ccw()
        else:
            self.turn_cw()

        self.draw()
        return True

    def focus_event(self, widget, event, tf):
        self._has_focus = tf
        self.draw()
        return True

    def _calc_action(self, x, y):
        ang_deg = np.degrees(np.arctan2(x - self.center[0],
                                        y - self.center[1]) + np.pi * 1.5)
        ang_deg = self.normalize_angle(ang_deg + self.ang_offset)
        if self.ang_invert:
            ang_deg = 360.0 - ang_deg

        self.angle_action(x, y, ang_deg)

    def draw(self):
        if self.surface is None:
            return
        cr = cairo.Context(self.surface)

        cr.select_font_face(self.fontname)
        cr.set_font_size(self.fontsize)

        # fill background
        wd, ht = self.dims
        cr.rectangle(0, 0, wd, ht)
        r, g, b = self.bg
        cr.set_source_rgba(r, g, b)
        cr.fill()

        r, g, b = self.fg
        cr.set_source_rgba(r, g, b)
        cr.set_line_width(2.0)

        cr.save()
        cx, cy = self.center
        cr.translate(cx, cy)
        cr.move_to(0, 0)

        # draw circle
        cradius = min(cx, cy)
        cradius *= 0.66
        cr.arc(0, 0, cradius, 0, 2 * np.pi)
        cr.fill()

        if self._has_focus:
            r, g, b = self.focus_fg
        else:
            r, g, b = (0.0, 0.0, 0.0)
        cr.set_source_rgba(r, g, b)
        cr.new_path()
        cr.set_line_width(2)
        cr.arc(0, 0, cradius, 0, 2 * np.pi)
        cr.stroke()
        cr.new_path()
        cr.set_line_width(1)

        if self.draw_scale:
            cr.new_path()
            cr.set_source_rgba(0.0, 0.0, 0.0)
            for tup in self.values:
                if len(tup) == 3:
                    label, value, theta = tup
                else:
                    value, theta = tup
                    label = str(value)
                if self.ang_invert:
                    theta = 360.0 - theta
                theta_pos = self.normalize_angle(theta + 90.0 - self.ang_offset)
                theta_rad = np.radians(theta_pos)
                a, b, wd, ht, i, j = cr.text_extents(label)
                crad2 = cradius + ht / 2.0
                if self.label_style == 0:
                    crad2 += wd

                # draw small filled dot as position marker
                cx, cy = (np.sin(theta_rad) * cradius,
                          np.cos(theta_rad) * cradius)
                cr.move_to(cx, cy)
                r, g, b = self.knob_fill
                cr.set_source_rgba(r, g, b)
                cr.arc(cx, cy, 2, 0, 2 * np.pi)
                cr.stroke_preserve()
                cr.fill()

                # draw label
                cx, cy = np.sin(theta_rad) * crad2, np.cos(theta_rad) * crad2
                cr.move_to(cx, cy)

                text_rad = np.arctan2(cx, cy)
                if self.label_style == 0:
                    text_rad = 0.0
                elif self.label_style == 1:
                    text_rad += np.pi
                elif self.label_style == 2:
                    text_rad += - np.pi / 2
                cr.save()
                cr.translate(cx, cy)
                cr.rotate(-text_rad)
                if self.label_style == 1:
                    cr.move_to(-wd / 2, 0)
                cr.show_text(label)
                #cr.rotate(text_rad)
                cr.restore()

            cr.new_path()

        cr.move_to(0, 0)
        theta = self.angle
        if self.ang_invert:
            theta = 360.0 - theta
        theta = self.normalize_angle(theta - self.ang_offset)

        cr.rotate(-np.radians(theta))

        # draw knob (pointer)
        r, g, b = self.knob_fg
        cr.set_source_rgba(r, g, b)
        crad2 = cradius
        cr.new_path()
        x1, y1, x2, y2 = -crad2, 0, crad2, 0
        cx1, cy1, cx2, cy2 = self.calc_vertexes(x1, y1, x2, y2,
                                                arrow_length=crad2)
        cr.move_to(x2, y2)
        cr.line_to(cx1, cy1)
        #cr.line_to(0, 0)
        cr.line_to(cx2, cy2)
        cr.close_path()
        r, g, b = self.knob_fg
        cr.set_source_rgba(r, g, b)
        cr.stroke_preserve()
        r, g, b = self.knob_fill
        cr.set_source_rgba(r, g, b)
        cr.fill()
        cr.move_to(0, 0)
        cr.arc(0, 0, abs(cx1 + cx2) * 2.1, 0, 2 * np.pi)
        cr.stroke_preserve()
        cr.fill()

        text = self.value_text
        if self.draw_value_pos == 1:
            r, g, b = self.bg
            cr.set_source_rgba(r, g, b)
            cr.move_to(0, 0)
            cr.show_text(text)

        cr.restore()

        if self.draw_value_pos == 2:
            a, b, wd, ht, i, j = cr.text_extents(text)
            r, g, b = self.fg
            cr.set_source_rgba(r, g, b)
            x, y = self.center
            cr.move_to(x - wd / 2, (y - cradius) * 0.5 + ht)
            cr.show_text(text)

        cr.move_to(0, 0)

        self.update_widget()

    def normalize_angle(self, ang_deg):
        ang_deg = np.fmod(ang_deg + 360.0, 360.0)
        return ang_deg

    def finalize_angle(self, ang_deg):
        self.angle = ang_deg
        self.draw()

    def get_angle(self):
        return self.angle

    def set_labels(self, val_ang_pairs):
        self.values = val_ang_pairs

        self.draw()

    def set_tracking(self, tf):
        self.tracking = tf

    def configure_event(self, widget, event):
        rect = widget.get_allocation()
        x, y, width, height = rect.x, rect.y, rect.width, rect.height

        self.dims = np.array((width, height))
        self.center = np.array((width / 2, height / 2))

        self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)

        self.draw()
        return True

    def update_widget(self):
        if self.surface is None:
            # window is not mapped/configured yet
            return

        win = self.get_window()
        if win is not None and self.surface is not None:
            wd, ht = self.dims

            self.queue_draw_area(0, 0, wd, ht)

    def draw_event(self, widget, cr):
        # redraw the screen from backing surface
        cr.set_source_surface(self.surface, 0, 0)

        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        return False

    def calc_vertexes(self, start_cx, start_cy, end_cx, end_cy,
                      arrow_length=10, arrow_degrees=0.35):

        angle = np.arctan2(end_cy - start_cy, end_cx - start_cx) + np.pi

        cx1 = end_cx + arrow_length * np.cos(angle - arrow_degrees)
        cy1 = end_cy + arrow_length * np.sin(angle - arrow_degrees)
        cx2 = end_cx + arrow_length * np.cos(angle + arrow_degrees)
        cy2 = end_cy + arrow_length * np.sin(angle + arrow_degrees)

        return (cx1, cy1, cx2, cy2)

    def angle_action(self, x, y, ang_deg):
        """Subclass overrides to provide custom behavior"""
        self._set_value(ang_deg)

    def turn_ccw(self):
        """Subclass overrides to provide custom behavior"""
        self._set_value(self.angle + self.turn_delta)

    def turn_cw(self):
        """Subclass overrides to provide custom behavior"""
        self._set_value(self.angle - self.turn_delta)

    def _set_value(self, ang_deg):
        """Subclass overrides to provide custom behavior"""
        ang_deg = self.normalize_angle(ang_deg)
        ang_deg = np.clip(ang_deg, self.min_ang_deg, self.max_ang_deg)
        self.value = ang_deg
        self.finalize_angle(ang_deg)

        if not self._dragging or self.tracking:
            self.emit("value-changed", self.value)

    def set_value(self, ang_deg):
        """Subclass overrides to provide custom behavior"""
        self._set_value(ang_deg)

    def get_value(self):
        """Subclass overrides to provide custom behavior"""
        return self.value


class ValueDial(Dial):

    __gtype_name__ = "ValueDial"

    def __init__(self):
        Dial.__init__(self)

        # for drawing value
        self.label_style = 1
        self.draw_value_pos = 2

        # setup axis orientation to match value
        self.ang_offset = 140.0
        self.ang_invert = True
        self.min_ang_deg = 0.0
        self.max_ang_deg = 260.0
        self.set_labels([("min", 0.0), ("max", 260.0)])

        self.min_val = 0.0
        self.max_val = 0.0
        self.inc_val = 0.0

    def angle_action(self, x, y, ang_deg):
        value = self._angle_to_value(ang_deg)
        self._set_value(value)

    def turn_ccw(self):
        ang_deg = np.clip(self.angle + self.turn_delta,
                          0.0, self.max_ang_deg)
        value = self._angle_to_value(ang_deg)
        self._set_value(value)

    def turn_cw(self):
        ang_deg = np.clip(self.angle - self.turn_delta,
                          0.0, self.max_ang_deg)
        value = self._angle_to_value(ang_deg)
        self._set_value(value)

    def _set_value(self, value):
        if value < self.min_val or value > self.max_val:
            raise ValueError("value '{}' is out of range".format(value))

        self.value = value
        self.value_text = "%.2f" % self.value

        ang_deg = self._value_to_angle(value)
        self.finalize_angle(ang_deg)

        if not self._dragging or self.tracking:
            self.emit("value-changed", self.value)

    def get_value(self):
        return self.value

    def _value_to_angle(self, value):
        # make angle match value
        rng = self.max_val - self.min_val
        pct = (value - self.min_val) / rng
        ang_deg = pct * self.max_ang_deg
        ang_deg = np.clip(ang_deg, 0.0, self.max_ang_deg)
        return ang_deg

    def _angle_to_value(self, ang_deg):
        # make value match angle
        pct = ang_deg / self.max_ang_deg
        rng = self.max_val - self.min_val
        value = self.min_val + pct * rng
        value = np.clip(value, self.min_val, self.max_val)
        return value

    def set_limits(self, min_val, max_val, inc_val):
        self.min_val = min_val
        self.max_val = max_val
        self.inc_val = inc_val

        pct = inc_val / (max_val - min_val)
        self.turn_delta = pct * self.max_ang_deg


class IndexDial(Dial):

    __gtype_name__ = "IndexDial"

    def __init__(self):
        Dial.__init__(self)

        self.idx = 0
        self.label_style = 1

    def angle_action(self, x, y, ang_deg):
        idx = self.best_index(ang_deg)
        self.set_index(idx)

    def turn_ccw(self):
        idx = self.idx - 1
        if idx < 0:
            if self.wrap:
                self.set_index(len(self.values) - 1)
        else:
            self.set_index(idx)

    def turn_cw(self):
        idx = self.idx + 1
        if idx >= len(self.values):
            if self.wrap:
                self.set_index(0)
        else:
            self.set_index(idx)

    def set_index(self, idx):
        idx = int(idx)
        if idx < 0 or idx >= len(self.values):
            raise ValueError("index '{}' is outside range 0-{}".format(idx,
                                                                       len(self.values)))
        self.idx = idx
        tup = self.values[idx]
        self.value = tup[0] if len(tup) == 2 else tup[1]
        self.value_text = str(self.value)

        self.angle = tup[-1]
        self.draw()

        if not self._dragging or self.tracking:
            self.emit("value-changed", idx)

    def get_index(self):
        return self.idx

    def get_value(self):
        return self.value

    def best_index(self, ang_deg):
        # find the index that is closest to the angle ang_deg
        angles = np.array([tup[-1] for tup in self.values])
        ang_deg = self.normalize_angle(ang_deg)
        angles = np.abs(angles - ang_deg)
        idx = np.argmin(angles)
        return idx


class FileSelection(object):

    def __init__(self, parent_w, action=Gtk.FileChooserAction.OPEN,
                 title="Select a file", all_at_once=False):
        # TODO: deprecate the functionality when all_at_once == False
        # and make the default to be True
        self.parent = parent_w
        self.all_at_once = all_at_once
        # Create a new file selection widget
        self.filew = Gtk.FileChooserDialog(title=title, action=action)
        self.filew.connect("destroy", self.close)
        if action == Gtk.FileChooserAction.SAVE:
            self.filew.add_buttons(Gtk.STOCK_SAVE, 1, Gtk.STOCK_CANCEL, 0)
        else:
            #self.filew.add_buttons(Gtk.STOCK_OPEN, 1, Gtk.STOCK_CANCEL, 0)
            self.filew.add_button("Open", 1)
            self.filew.add_button("Cancel", 0)
        self.filew.set_default_response(1)
        self.filew.set_select_multiple(True)
        self.filew.connect("response", self.file_ok_sel)

        # Connect the cancel_button to destroy the widget
        #self.filew.cancel_button.connect("clicked", self.close)

    def popup(self, title, callfn, initialdir=None, filename=None):
        """Let user select and load file."""
        self.cb = callfn
        self.filew.set_title(title)
        if initialdir:
            self.filew.set_current_folder(initialdir)

        if filename:
            #self.filew.set_filename(filename)
            self.filew.set_current_name(filename)

        self.filew.show()
        # default size can be enormous
        #self.filew.resize(800, 600)

    # Get the selected filename
    def file_ok_sel(self, w, rsp):
        self.close(w)
        if rsp == 0:
            return

        gdk_paths = self.filew.get_files()
        paths = [gdk_path.get_path() for gdk_path in gdk_paths]
        if self.all_at_once:
            self.cb(paths)
        else:
            for path in paths:
                self.cb(path)

    def close(self, widget):
        self.filew.hide()


class DirectorySelection(FileSelection):
    """Handle directory selection dialog."""
    def __init__(self, parent_w):
        super(DirectorySelection, self).__init__(
            parent_w, action=Gtk.FileChooserAction.SELECT_FOLDER,
            title="Select a directory")

    def popup(self, title, callfn, initialdir=None):
        """Let user select a directory."""
        super(DirectorySelection, self).popup(title, callfn, initialdir)


class Timer(Callback.Callbacks):
    """Abstraction of a GUI-toolkit implemented timer."""

    def __init__(self, duration=0.0):
        """Create a timer set to expire after `duration` sec.
        """
        super(Timer, self).__init__()

        self.duration = duration
        # For storing aritrary data with timers
        self.data = Bunch.Bunch()

        self._timer = None
        self.start_time = 0.0
        self.deadline = 0.0

        for name in ('expired', 'canceled'):
            self.enable_callback(name)

    def start(self, duration=None):
        """Start the timer.  If `duration` is not None, it should
        specify the time to expiration in seconds.
        """
        if duration is None:
            duration = self.duration

        self.set(duration)

    def set(self, duration):

        self.stop()

        self.start_time = time.time()
        self.deadline = self.start_time + duration
        # Gtk timer set in milliseconds
        time_ms = int(duration * 1000.0)
        self._timer = GObject.timeout_add(time_ms, self._redirect_cb)

    def _redirect_cb(self):
        self._timer = None
        self.make_callback('expired')

    def is_set(self):
        return self._timer is not None

    def cond_set(self, time_sec):
        if not self.is_set():
            # TODO: probably a race condition here
            self.set(time_sec)

    def elapsed_time(self):
        return time.time() - self.start_time

    def time_left(self):
        return max(0.0, self.deadline - time.time())

    def get_deadline(self):
        return self.deadline

    def stop(self):
        try:
            if self._timer is not None:
                GObject.source_remove(self._timer)
        except Exception:
            pass
        self._timer = None

    def cancel(self):
        """Cancel this timer.  If the timer is not running, there
        is no error.
        """
        self.stop()
        self.make_callback('canceled')

    clear = cancel


def combo_box_new_text():
    liststore = Gtk.ListStore(GObject.TYPE_STRING)
    combobox = ComboBox()
    combobox.set_model(liststore)
    cell = Gtk.CellRendererText()
    combobox.pack_start(cell, True)
    combobox.add_attribute(cell, 'text', 0)
    return combobox


def get_scroll_info(dx, dy):
    """
    Returns the (degrees, direction) of a scroll motion Gtk event.
    """
    # gtk4 scrolls in the opposite direction vertically from what
    # we are used to doing
    dy = -dy
    # we have a trackpad or some device that reports pixel deltas
    delta = math.sqrt(dx ** 2 + dy ** 2)
    if dy < 0:
        delta = -delta

    ang_rad = math.atan2(dy, dx)
    direction = math.degrees(ang_rad) - 90.0
    direction = math.fmod(direction + 360.0, 360.0)

    # TODO: is this accurate?--NOT TESTED
    num_degrees = delta / 8.0

    return (num_degrees, direction)


def get_icon(iconpath, size=None, adjust_width=True):
    if size is not None:
        wd, ht = size
    else:
        wd, ht = 24, 24

    if adjust_width:
        # TODO
        pixbuf = pixbuf_new_from_file(iconpath)
        # get "natural" dimensions
        _w, _h = pixbuf.get_width(), pixbuf.get_height()
        aspect = _w / _h
        wd = int(wd * aspect)

    pixbuf = pixbuf_new_from_file_at_size(iconpath, wd, ht)
    return pixbuf


def get_font(font_family, point_size):
    font_family = font_asst.resolve_alias(font_family, font_family)
    font = Pango.FontDescription('%s %d' % (font_family, point_size))
    return font


def load_font(font_name, font_file):
    # TODO!
    ## raise ValueError("Loading fonts dynamically is an unimplemented"
    ##                  " feature for gtk4 back end")
    return font_name


# def pixbuf_new_from_xpm_data(xpm_data):
#     xpm_data = bytes('\n'.join(xpm_data))
#     return GdkPixbuf.Pixbuf.new_from_xpm_data(xpm_data)


# def pixbuf_new_from_array(data, rgbtype, bpp):
#     # NOTE: there is a bug in gtk4 with pixbuf_new_from_array()
#     # See: http://stackoverflow.com/questions/24062779/how-to-correctly-covert-3d-array-into-continguous-rgb-bytes/24070152#24070152
#     #return GdkPixbuf.Pixbuf.new_from_array(data, rgbtype, bpp)

#     height, width, depth = data.shape
#     pixl = GdkPixbuf.PixbufLoader.new_with_type('pnm')
#     # P6 is the magic number of PNM format,
#     # and 255 is the max color allowed
#     pixl.write((b"P6 %d %d 255 " % (width, height)) + data.tobytes(order='C'))
#     pix = pixl.get_pixbuf()
#     pixl.close()
#     return pix


# def pixbuf_new_from_data(rgb_buf, rgbtype, hasAlpha, bpp, dawd, daht, stride):
#     return GdkPixbuf.Pixbuf.new_from_data(rgb_buf, rgbtype, hasAlpha, bpp,
#                                           dawd, daht, stride, None, None)


def pixbuf_new_from_file_at_size(foldericon, width, height):
    return GdkPixbuf.Pixbuf.new_from_file_at_size(foldericon, width, height)


def pixbuf_new_from_file(file_path):
    return GdkPixbuf.Pixbuf.new_from_file(file_path)


def make_cursor(widget, iconpath, x, y, size=None):
    if size is None:
        size = (16, 16)
    fallback = Gdk.Cursor.new_from_name("crosshair")
    pixbuf = pixbuf_new_from_file_at_size(iconpath, size[0], size[1])
    texture = Gdk.Texture.new_for_pixbuf(pixbuf)
    cursor = Gdk.Cursor.new_from_texture(texture, x, y, fallback)
    return cursor


def modify_bg(widget, color):
    context = widget.get_style_context()
    if color is not None:
        context.add_class("custom_bg")
        css_data = "*.custom_bg { background-image: none; background-color: %s; }" % (color)
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(css_data.encode())
        context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)
    else:
        context.remove_class("custom_bg")


def set_border_width(widget, pix):
    if hasattr(widget, 'set_margin_start'):
        widget.set_margin_start(pix)
        widget.set_margin_end(pix)
        widget.set_margin_top(pix)
        widget.set_margin_bottom(pix)
    else:
        raise ValueError("No way to set border width on this widget")


def set_default_style():
    style_provider = Gtk.CssProvider()

    module_home = os.path.split(sys.modules[__name__].__file__)[0]
    gtk_css = os.path.join(module_home, 'gtk_css')
    with open(gtk_css, 'rb') as css_f:
        css_data = css_f.read()

    try:
        style_provider.load_from_data(css_data)

        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    except Exception:
        pass



"""

Usage:
  $ bokeh serve example1.py

you will see something like this in the output:
2017-11-29 17:00:48,368 Starting Bokeh server version 0.12.7 (running on Tornado 4.3)
2017-11-29 17:00:48,371 Bokeh app running at: http://localhost:5006/example1
2017-11-29 17:00:48,371 Starting Bokeh server with process id: 14214

Visit the URL with a browser to interact with the GUI.  Enter a valid
FITS file path into the box labeled "File:" and press Enter.
"""

import os
import tempfile
#import io, base64

from bokeh.layouts import column
from bokeh.plotting import figure, curdoc
from bokeh.models import TextInput, RadioButtonGroup, Div
#from bokeh.models.widgets import FileInput

from ginga.web.bokehw.ImageViewBokeh import CanvasView
from ginga.misc import log
from ginga import AstroImage
from ginga.util.loader import load_data


def main():

    # "bokeh serve" chokes on --log
    log_file = os.path.join(tempfile.gettempdir(), "example1.log")
    logger = log.get_logger("ginga", level=20, log_file=log_file)

    # viewer size
    wd, ht = 600, 600
    TOOLS = ""

    # create a new plot with default tools, using figure
    fig = figure(x_range=[0, wd], y_range=[0, ht],
                 width=wd, height=ht,
                 toolbar_location="below", toolbar_sticky=False,
                 tools=TOOLS, output_backend="webgl")
    viewer = CanvasView(logger=logger)
    viewer.set_figure(fig)
    viewer.enable_autocuts('on')
    viewer.set_color_map('rainbow3')
    viewer.set_autocut_params('zscale')
    viewer.enable_autozoom('on')
    viewer.set_zoom_algorithm('rate')
    viewer.set_zoomrate(1.4)
    viewer.show_pan_mark(True)
    viewer.show_mode_indicator(True)

    bd = viewer.get_bindings()
    bd.enable_all(True)
    bd.set_mode(viewer, 'pan', mode_type='locked')

    def load_file(path):
        image = load_data(path, logger=logger)
        viewer.set_image(image)

    def load_file_cb(attr, old_val, new_val):
        try:
            img = AstroImage.AstroImage(logger=logger)
            img.load_file(new_val.strip())
            viewer.set_image(img)
        except Exception as e:
            logger.error("problem loading file: {}".format(e), exc_info=True)

    # def upload_file_cb(attr, old_val, new_val):
    #     try:
    #         buf_bytes = base64.b64decode(new_val)
    #         io_f = io.BytesIO(buf_bytes)
    #         img = AstroImage.AstroImage(logger=logger)
    #         with fits.open(io_f, 'readonly') as in_f:
    #             img.load_hdu(in_f[0])
    #         viewer.set_image(img)
    #     except Exception as e:
    #         logger.error("problem loading file: {}".format(e), exc_info=True)

    labels = ["Pan", "Rotate", "Contrast", "Dist", "Cuts", "Cmap"]

    def mode_change_cb(attr_name, old_val, new_val):
        mode = labels[new_val].lower()
        bd.set_mode(viewer, mode, mode_type='locked')

    readout = Div(text="")

    def cursor_cb(viewer, button, data_x, data_y):
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + viewer.data_off),
                                    int(data_y + viewer.data_off))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        readout.text = text

    viewer.add_callback('cursor-changed', cursor_cb)

    mode = RadioButtonGroup(labels=labels, active=0)
    mode.on_change('active', mode_change_cb)

    # add a entry widget and configure with the call back
    path_w = TextInput(value=os.environ['HOME'], title="File:")
    path_w.on_change('value', load_file_cb)

    layout = column(fig, readout, mode, path_w)
    curdoc().add_root(layout)


main()



#! /usr/bin/env python
#
# example2.py -- Simple, configurable FITS viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import sys

import ginga.toolkit as ginga_toolkit
from ginga import colors
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.canvas import render
from ginga.misc import log
from ginga.util.loader import load_data
from ginga.locale.localize import _tr


class FitsViewer(object):

    def __init__(self, logger, render='widget'):
        self.logger = logger
        self.drawcolors = colors.get_colors()
        self.dc = get_canvas_types()

        from ginga.gw import Widgets, Viewers, GwHelp

        self.app = Widgets.Application(logger=logger)
        self.app.add_callback('shutdown', self.quit)
        if hasattr(Widgets, 'Page'):
            self.page = Widgets.Page(_tr("Ginga example2"))
            self.app.add_window(self.page)
            self.top = Widgets.TopLevel(_tr("Ginga example2"))
            self.page.add_dialog(self.top)
        else:
            self.top = Widgets.TopLevel(_tr("Ginga example2"))
            self.app.add_window(self.top)
        self.top.add_callback('close', self.closed)

        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(1)

        fi = Viewers.CanvasView(logger=logger, render=render)
        fi.enable_autocuts('on')
        fi.set_autocut_params('zscale')
        fi.enable_autozoom('on')
        fi.set_zoom_algorithm('rate')
        fi.set_zoomrate(1.4)
        fi.show_pan_mark(True)
        fi.show_focus_indicator(True)
        fi.set_enter_focus(True)
        fi.set_callback('drag-drop', self.drop_file_cb)
        fi.set_callback('cursor-changed', self.cursor_cb)
        fi.set_bg(0.2, 0.2, 0.2)
        fi.ui_set_active(True)
        self.fitsimage = fi

        bd = fi.get_bindings()
        bd.enable_all(True)

        # add a color bar
        #fi.private_canvas.add(self.dc.ColorBar(side='bottom', offset=10))

        # add little mode indicator that shows modal states in
        # lower left hand corner
        fi.private_canvas.add(self.dc.ModeIndicator(corner='ur', fontsize=14))
        # little hack necessary to get correct operation of the mode indicator
        # in all circumstances
        bm = fi.get_bindmap()
        bm.add_callback('mode-set', lambda *args: fi.redraw(whence=3))

        # canvas that we will draw on
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype('rectangle', color='lightblue', coord='data')
        canvas.set_surface(fi)
        self.canvas = canvas
        # add canvas to view
        fi.get_canvas().add(canvas)
        canvas.ui_set_active(True)
        canvas.register_for_cursor_drawing(fi)
        canvas.add_callback('draw-event', self.draw_cb)

        self.drawtypes = canvas.get_drawtypes()
        self.drawtypes.sort()

        fi.set_desired_size(512, 512)
        iw = Viewers.GingaScrolledViewerWidget(viewer=fi)
        vbox.add_widget(iw, stretch=1)

        self.readout = Widgets.Label("")
        vbox.add_widget(self.readout, stretch=0)

        wdrawtype = Widgets.ComboBox()
        for name in self.drawtypes:
            wdrawtype.append_text(name)
        index = self.drawtypes.index('rectangle')
        wdrawtype.set_index(index)
        wdrawtype.set_tooltip("choose shape to draw")
        wdrawtype.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wdrawtype = wdrawtype

        wdrawcolor = Widgets.ComboBox()
        for name in self.drawcolors:
            wdrawcolor.append_text(name)
        index = self.drawcolors.index('lightblue')
        wdrawcolor.set_index(index)
        wdrawcolor.set_tooltip("choose line color")
        wdrawcolor.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wdrawcolor = wdrawcolor

        wfillcolor = Widgets.ComboBox()
        for name in self.drawcolors:
            wfillcolor.append_text(name)
        index = self.drawcolors.index('seagreen')
        wfillcolor.set_index(index)
        wfillcolor.set_tooltip("choose fill color")
        wfillcolor.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wfillcolor = wfillcolor

        wwidth = Widgets.ComboBox()
        for val in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
            wwidth.append_text(str(val))
        wwidth.set_index(1)
        wwidth.set_tooltip("set line width")
        wwidth.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wwidth = wwidth

        walpha = Widgets.SpinBox(dtype=float)
        walpha.set_limits(0.0, 1.0, incr_value=0.1)
        walpha.set_value(1.0)
        walpha.set_decimals(2)
        walpha.set_tooltip("set alpha value for line")
        walpha.add_callback('value-changed', lambda w, val: self.set_drawparams())
        self.walpha = walpha

        wfillalpha = Widgets.SpinBox(dtype=float)
        wfillalpha.set_limits(0.0, 1.0, incr_value=0.1)
        wfillalpha.set_value(0.0)
        wfillalpha.set_decimals(2)
        wfillalpha.set_tooltip("set alpha value for fill")
        wfillalpha.add_callback('value-changed', lambda w, val: self.set_drawparams())
        self.wfillalpha = wfillalpha

        renderers = ['cairo', 'pil', 'agg', 'opencv']
        if ginga_toolkit.family.startswith('qt'):
            renderers.extend(['qt', 'vqt'])
        wrender = Widgets.ComboBox()
        for name in renderers:
            wrender.append_text(name)
        wrender.set_index(0)
        wrender.set_tooltip("choose renderer backend")
        wrender.add_callback('activated', lambda w, idx: self.set_renderer())
        self.wrender = wrender

        wclear = Widgets.Button(_tr("Clear Canvas"))
        wclear.add_callback('activated', lambda w: self.clear_canvas())
        wopen = Widgets.Button(_tr("Open File"))
        wopen.add_callback('activated', lambda w: self.open_file())
        wquit = Widgets.Button(_tr("Quit"))
        wquit.add_callback('activated', lambda w: self.quit())

        hbox = Widgets.HBox()
        hbox.set_border_width(2)
        hbox.set_spacing(4)

        for w in (wopen, wdrawtype, wdrawcolor, wwidth,
                  Widgets.Label(_tr('Alpha') + ':'), walpha):
            hbox.add_widget(w, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)

        vbox.add_widget(hbox, stretch=0)

        hbox = Widgets.HBox()
        hbox.set_border_width(2)
        hbox.set_spacing(4)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        for w in (wrender, wfillcolor, wfillalpha, wclear, wquit):
            hbox.add_widget(w, stretch=0)

        vbox.add_widget(hbox, stretch=0)
        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        hbox.set_spacing(4)
        btn1 = Widgets.RadioButton(_tr("Draw"))
        btn1.set_state(mode == 'draw')
        btn1.add_callback('activated', lambda w, val: self.set_mode_cb('draw', val))
        btn1.set_tooltip(_tr("Choose this to draw on the canvas"))
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton(_tr("Edit"), group=btn1)
        btn2.set_state(mode == 'edit')
        btn2.add_callback('activated', lambda w, val: self.set_mode_cb('edit', val))
        btn2.set_tooltip(_tr("Choose this to edit things on the canvas"))
        hbox.add_widget(btn2)

        btn3 = Widgets.RadioButton(_tr("Pick"), group=btn1)
        btn3.set_state(mode == 'pick')
        btn3.add_callback('activated', lambda w, val: self.set_mode_cb('pick', val))
        btn3.set_tooltip(_tr("Choose this to pick things on the canvas"))
        hbox.add_widget(btn3)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(hbox, stretch=0)

        self.top.set_widget(vbox)

        self.fs = None
        if hasattr(GwHelp, 'FileSelection'):
            self.fs = GwHelp.FileSelection(self.top.get_widget())

    def set_drawparams(self):
        index = self.wdrawtype.get_index()
        kind = self.drawtypes[index]
        index = self.wdrawcolor.get_index()
        linewidth = int(self.wwidth.get_text())
        alpha = self.walpha.get_value()
        fillalpha = self.wfillalpha.get_value()

        params = {'color': self.drawcolors[index],
                  'linewidth': linewidth,
                  'alpha': alpha,
                  }
        if kind in ('circle', 'rectangle', 'polygon', 'triangle',
                    'righttriangle', 'ellipse', 'square', 'box',
                    'squarebox', 'text'):
            params['fill'] = True
            index = self.wfillcolor.get_index()
            params['fillcolor'] = self.drawcolors[index]
            params['fillalpha'] = fillalpha

        self.canvas.set_drawtype(kind, **params)

    def clear_canvas(self):
        self.canvas.delete_all_objects()

    def set_renderer(self):
        name = self.wrender.get_text()
        klass = render.get_render_class(name)
        renderer = klass(self.fitsimage)
        self.fitsimage.set_renderer(renderer)
        self.fitsimage.redraw(whence=0)

    def load_file(self, filepath):
        image = load_data(filepath, logger=self.logger)
        self.fitsimage.set_image(image)
        self.top.set_title(filepath)

    def open_file(self):
        self.fs.popup(_tr("Open FITS file"), self.load_file)

    def drop_file_cb(self, viewer, paths):
        filename = paths[0]
        self.load_file(filename)

    def cursor_cb(self, viewer, button, data_x, data_y):
        """This gets called when the data position relative to the cursor
        changes.
        """
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + viewer.data_off),
                                    int(data_y + viewer.data_off))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning(_tr("Bad coordinate conversion") + ": %s" % (
                str(e)))
            ra_txt = _tr('BAD WCS')
            dec_txt = _tr('BAD WCS')

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.set_text(text)

    def set_mode_cb(self, mode, tf):
        self.logger.info("canvas mode changed (%s) %s" % (mode, tf))
        if not (tf is False):
            self.canvas.set_draw_mode(mode)
        return True

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        obj.add_callback('pick-down', self.pick_cb, 'down')
        obj.add_callback('pick-up', self.pick_cb, 'up')
        obj.add_callback('pick-move', self.pick_cb, 'move')
        obj.add_callback('pick-hover', self.pick_cb, 'hover')
        obj.add_callback('pick-enter', self.pick_cb, 'enter')
        obj.add_callback('pick-leave', self.pick_cb, 'leave')
        obj.add_callback('pick-key', self.pick_cb, 'key')
        obj.pickable = True
        obj.add_callback('edited', self.edit_cb)

    def pick_cb(self, obj, canvas, event, pt, ptype):
        self.logger.info("pick event '%s' with obj %s at (%.2f, %.2f)" % (
            ptype, obj.kind, pt[0], pt[1]))
        return True

    def edit_cb(self, obj):
        self.logger.info("object %s has been edited" % (obj.kind))
        return True

    def closed(self, w):
        self.logger.info("Top window closed.")
        self.top = None
        sys.exit()

    def quit(self, *args):
        self.logger.info("Attempting to shut down the application...")
        if self.top is not None:
            self.top.close()
        sys.exit()


def main(options, args):

    logger = log.get_logger("example2", options=options)

    if options.toolkit is None:
        logger.error("Please choose a GUI toolkit with -t option")

    # decide our toolkit, then import
    ginga_toolkit.use(options.toolkit)

    rw = 'opengl' if options.renderer == 'opengl' else 'widget'
    viewer = FitsViewer(logger, render=rw)

    if options.renderer is not None and options.renderer != 'opengl':
        render_class = render.get_render_class(options.renderer)
        viewer.fitsimage.set_renderer(render_class(viewer.fitsimage))

    viewer.top.resize(700, 540)

    if len(args) > 0:
        viewer.load_file(args[0])

    viewer.top.show()
    viewer.top.raise_()

    try:
        app = viewer.top.get_app()
        app.mainloop()

    except KeyboardInterrupt:
        print("Terminating viewer...")
        if viewer.top is not None:
            viewer.top.close()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt)")
    argprs.add_argument("-r", "--renderer", dest="renderer", metavar="NAME",
                        default=None,
                        help="Choose renderer (pil|agg|opencv|cairo|qt)")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    main(options, args)



#
# plotaide.py -- Utility class for plotting using Ginga widgets.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time

import numpy as np

from ginga.misc import Bunch, Callback, Settings
from ginga.canvas import transform
from ginga.modes import modeinfo

from .plot_mode import PlotMode
modeinfo.add_mode(PlotMode)


class PlotAide(Callback.Callbacks):
    """
    PlotAide is a class to use a Ginga viewer as a speedy, interactive
    X/Y line plot viewer.
    """

    def __init__(self, viewer, settings=None):
        """PlotAide constructor.

        Parameters
        ----------
        viewer : subclass instance of `~ginga.ImageView.ImageViewBase`
            The ginga viewer that will be used to view the plot

        settings : `~ginga.misc.Settings.SettingGroup` or `None`
            Settings for the PlotAide.  If None, then a new SettingGroup
            will be created internally.
        """
        super(PlotAide, self).__init__()

        self.logger = viewer.get_logger()

        self.norm_bg = viewer.get_bg()
        self.norm_fg = viewer.get_fg()
        self.axis_bg = 'gray90'
        self.grid_fg = 'gray30'
        # bbox for the plot area (a subset of the ginga viewer window)
        # is (LL, LR, UR, UL) in typical window coordinates. This will
        # be adjusted later in the resize callbacks
        self.bbox = np.array([(0, 0), (0, 0), (0, 0), (0, 0)])
        self.llur = (0, 0, 0, 0)

        if settings is None:
            settings = Settings.SettingGroup(name='plotaide',
                                             logger=self.logger)
        self.settings = settings
        self.settings.set_defaults(autoaxis_x='on', autoaxis_y='on',
                                   tweak_lim_pct=0.05)

        # internal flags used for synchronization
        self._panning_x = False
        self._scaling_y = False
        self._scaling_x = False
        self._adjusting = False

        self.viewer = viewer
        # configure the ginga viewer for a plot
        vi = viewer
        vi.set_background(self.norm_bg)

        # turn off sanity check on scaling, since scaling can get
        # wild with plots
        t_ = vi.get_settings()
        t_.set(sanity_check_scale=False)

        # add special plot transform, used for the plot area
        self.plot_tr = transform.ScaleOffsetTransform()
        self.plot_tr.set_plot_scaling(1.0, 1.0, 0, 0)
        vi.tform['data_to_plot'] = (vi.tform['data_to_native'] + self.plot_tr)

        vi.enable_autozoom('off')
        vi.ui_set_active(True)
        vi.set_enter_focus(True)
        vi.add_callback('configure', lambda *args: self.adjust_resize_cb())
        vi.add_callback('transform', lambda *args: self.adjust_resize_cb())

        # disable normal user interaction, except flip
        bd = vi.get_bindings()
        bd.enable_all(False)
        bd.enable_flip(True)
        self.bd = bd

        # enables plot mode specific bindings
        mode = self.bd.get_mode_obj('plot')
        mode.set_aide_ref(self)
        self.bd.set_mode(vi, 'plot', mode_type='locked')

        for name in ['plot-zoom-x', 'plot-zoom-y', 'pan']:
            self.enable_callback(name)
        self.add_callback('plot-zoom-x', self.plot_zoom_x_cb)
        self.add_callback('plot-zoom-y', self.plot_zoom_y_cb)
        self.add_callback('pan', self.pan_cb)

        self.settings.get_setting('autoaxis_x').add_callback('set',
                                                             self._autoaxis_x_cb)
        self.settings.get_setting('autoaxis_y').add_callback('set',
                                                             self._autoaxis_y_cb)

        settings = vi.get_settings()
        settings.get_setting('pan').add_callback('set', self._pan_cb)

        # get our canvas
        canvas = vi.get_canvas()
        self.dc = canvas.get_draw_classes()
        self.canvas = self.dc.DrawingCanvas()
        canvas.add(self.canvas)
        self.canvas.register_for_cursor_drawing(vi)
        self.canvas.set_draw_mode('pick')
        self.canvas.ui_set_active(True, viewer=vi)

        # holds our individual XY plots
        self.plots = Bunch.Bunch()
        # holds other plot elements like PlotBG, XAxis, YAxis, PlotTitle, etc.
        self.plot_decor_l = []
        self.plot_decor_d = {}
        self.titles = [None, None]

    def get_settings(self):
        return self.settings

    def add_plot_decor(self, plotable):
        """Add a "plot accessory" (``PlotBG``, ``PlotTitle``, ``XAxis``, etc)

        Parameters
        ----------
        plotable : `~ginga.canvas.types.plots.CompoundObject` instance
            A plot accessory like a PlotBG, PlotTitle, XAxis, YAxis, etc.
        """
        plotable.register_decor(self)
        self.plot_decor_d[plotable.kind] = plotable
        self.plot_decor_l.append(plotable)
        self.canvas.add(plotable)

        # set callbacks for tool tips
        if plotable.kind == 'plot_bg':
            plotable.set_callback('pick-down', self.show_tt_cb, 'down')
            plotable.set_callback('pick-move', self.show_tt_cb, 'move')
            plotable.set_callback('pick-up', self.show_tt_cb, 'up')

    def get_plot_decor(self, kind):
        """Get a "plot accessory"

        Parameters
        ----------
        kind : str
            One of ('axis_x', 'axis_y', 'plot_bg', 'plot_title')

        Returns
        -------
        plotable : `~ginga.canvas.types.layer.CompoundObject` instance
            The plot accessory
        """
        return self.plot_decor_d[kind]

    def add_plot(self, plot):
        """Add a XY plot

        Parameters
        ----------
        plot : `~ginga.canvas.types.plots.XYPlot` instance
            A plot to add to the viewer
        """
        name = plot.name
        self.plots[name] = plot

        self.canvas.add(plot, tag=name)
        plot_bg = self.plot_decor_d.get('plot_bg', None)
        self.canvas.raise_object(plot, aboveThis=plot_bg)

        for plotable in self.plot_decor_l:
            plotable.add_plot(self.viewer, plot)

        limits = self.get_limits('data')
        x1, y1 = limits[0][:2]
        x2, y2 = limits[1][:2]

        self.viewer.set_limits(limits)
        self.viewer.set_pan((x1 + x2) * 0.5, (y1 + y2) * 0.5)

        self.update_plots()

    def delete_plot(self, plot):
        """Delete a XY plot

        Parameters
        ----------
        plot : `~ginga.canvas.types.plots.XYPlot` instance
            The plot to delete from the viewer
        """
        name = plot.name
        del self.plots[name]

        self.canvas.delete_object_by_tag(name)

        for plotable in self.plot_decor_l:
            plotable.delete_plot(self.viewer, plot)

        self.update_plots()

    def get_plot(self, name):
        """Get a XY plot

        Raises a `KeyError` if the requested plot is not present.

        Parameters
        ----------
        name : `str`
            The name of the plot

        Returns
        -------
        plot : `~ginga.canvas.types.plots.XYPlot` instance
            The requested plot
        """
        return self.plots[name]

    def update_plots(self):
        """Update the plots visibly after some activity, like plotting
        new data, or changing the data of a plot.
        """
        st = time.time()

        self.adjust_view()

        et = time.time()
        self.logger.debug("%.5f sec to update plot" % (et - st))

    def update_elements(self):
        """Mostly internal routine used to update the plot accessories
        after changes in region shown or data updates.
        """
        with self.viewer.suppress_redraw:
            for plotable in self.plot_decor_l:
                plotable.update_elements(self.viewer)

            self.viewer.redraw(whence=3)

    def update_resize(self):
        """Mostly internal routine called after the viewer's window size
        has changed.
        """
        # upon a resize, at first we redefine the plot bbox to encompass
        # the entire viewer area
        dims = self.viewer.get_window_size()
        wd, ht = dims[:2]
        # bbox is (LL, LR, UR, UL) in typical window coordinates
        self.bbox = np.array([(0, ht), (wd, ht), (wd, 0), (0, 0)])

        with self.viewer.suppress_redraw:
            # update all non-plot elements: axes, title, bg, etc
            # after this the plot bbox will have been redefined
            for plotable in self.plot_decor_l:
                plotable.update_bbox(self.viewer, dims)

            # update the plot transform based on the new plot bbox
            xy_lim = self.recalc_transform(dims)

            # second pass to allow non-plot elements to position their
            # components now that plox bbox is known
            for plotable in self.plot_decor_l:
                plotable.update_resize(self.viewer, dims, xy_lim)

            # now update all plots, due to the new transform
            for plot_src in self.plots.values():
                plot_src.update_resize(self.viewer, dims)

            # finally, redraw everything
            self.viewer.redraw(whence=3)

    def update_plot_bbox(self, x_lo=None, x_hi=None, y_lo=None, y_hi=None):
        """Mostly internal routine called by plot accessories to iteratively
        define the bounding box of the plot within the viewer plane.
        """
        _x_lo, _x_hi = self.bbox.T[0].min(), self.bbox.T[0].max()
        _y_lo, _y_hi = self.bbox.T[1].min(), self.bbox.T[1].max()

        if x_lo is not None:
            _x_lo = x_lo
        if x_hi is not None:
            _x_hi = x_hi
        if y_lo is not None:
            _y_lo = y_lo
        if y_hi is not None:
            _y_hi = y_hi

        # just in case caller got any of these mixed up
        x_lo, x_hi = min(_x_lo, _x_hi), max(_x_lo, _x_hi)
        y_lo, y_hi = min(_y_lo, _y_hi), max(_y_lo, _y_hi)

        self.bbox = np.array([(_x_lo, _y_hi), (_x_hi, _y_hi),
                              (_x_hi, _y_lo), (_x_lo, _y_lo)])

    def recalc_transform(self, dims):
        """Mostly internal routine used to redefine the special transform
        that maps the data to the plot area.
        """
        x_lo, x_hi = self.bbox.T[0].min(), self.bbox.T[0].max()
        y_lo, y_hi = self.bbox.T[1].min(), self.bbox.T[1].max()

        wd, ht = dims[:2]

        x_pct = (x_hi - x_lo) / wd
        y_pct = (y_hi - y_lo) / ht

        self.plot_tr.set_plot_scaling(x_pct, y_pct, x_lo, y_lo)
        self.llur = (x_lo, y_lo, x_hi, y_hi)

        return self.llur

    def set_limits(self, limits):
        self.viewer.set_limits(limits)

    def get_limits(self, lim_type):
        """Return the limits of this plot.

        Parameters
        ----------
        lim_type : str
            'data' or 'plot'

        Returns
        -------
        limits : array of [(x_min, y_min), (x_max, y_max)]
            Limits of full data or visible part of plot, depending on `lim_type`

        """
        x_vals = []
        y_vals = []
        for key, src in self.plots.items():
            limits = src.get_limits(lim_type)
            x_vals.append(limits.T[0])
            y_vals.append(limits.T[1])
        x_vals = np.array(x_vals).flatten()
        y_vals = np.array(y_vals).flatten()
        if len(x_vals) == 0:
            return np.array([(0.0, 0.0), (0.0, 0.0)])

        return np.array([(x_vals.min(), y_vals.min()),
                         (x_vals.max(), y_vals.max())])

    def tweak_limits(self, lo, hi):
        """Ensure that we can see the plotted lines by tweaking the
        limits just a tad to be just outside their range.
        """
        spread = hi - lo
        margin_pct = self.settings.get('tweak_lim_pct', 0.05)
        lo -= spread * margin_pct
        hi += spread * margin_pct
        return lo, hi

    def setup_standard_frame(self, title='', x_title=None, y_title=None,
                             num_x_labels=4, num_y_labels=4,
                             warn_y=None, alert_y=None):
        """A convenience function for setting up a typical plot.

        This includes a plot background, a title bar with keys, and X and Y
        axes.
        """
        from ginga.canvas.types import plots as gplots

        self.titles = [x_title, y_title]

        p_bg = gplots.PlotBG(linewidth=2, warn_y=warn_y, alert_y=alert_y)
        self.add_plot_decor(p_bg)

        p_title = gplots.PlotTitle(title=title)
        self.add_plot_decor(p_title)

        p_x_axis = gplots.XAxis(title=x_title, num_labels=num_x_labels)
        self.add_plot_decor(p_x_axis)

        p_y_axis = gplots.YAxis(title=y_title, num_labels=num_y_labels)
        self.add_plot_decor(p_y_axis)

    def get_axes_titles(self):
        """A function used by Axis subclasses to find their title.
        (title can shift depending on whether the viewer has swapped axes)
        """
        x, y = 0, 1
        flip = self.viewer.get_transforms()
        if flip[2]:
            x, y = y, x
        return dict(axis_x=self.titles[x], axis_y=self.titles[y])

    def configure_scrollbars(self, sw):
        """If a scroll widget is used with this plot viewer, then the
        widget can be passed to this method to configure the scrollbars
        to turn on and off as necessary depending on auto axis adjustments.
        """
        self.settings.get_setting('autoaxis_x').add_callback('set',
                                                             self.update_horz_scrollbar_cb, sw)
        self.settings.get_setting('autoaxis_y').add_callback('set',
                                                             self.update_vert_scrollbar_cb, sw)

        if self.settings['autoaxis_x'] in ('on', 'pan'):
            horz = 'off'
        else:
            horz = 'on'
        vert = 'off' if self.settings['autoaxis_y'] == 'on' else 'off'
        sw.scroll_bars(horizontal=horz, vertical=vert)

    def adjust_view(self):
        """Called after a pan, zoom or change of data, in order to
        autoscale X or Y axes, autopan the X axis and update the non-plot
        accessories (axes tick labels, etc).
        """
        if self._adjusting:
            return
        self._adjusting = True
        try:
            limits = self.get_limits('data')

            with self.viewer.suppress_redraw:
                # set limits to combined data limits of all plots in this
                # viewer
                self.viewer.set_limits(limits)

                # X limits may be adjusted by autoscaling X axis or
                # autopanning to most recent data
                if not self._scaling_x and self.settings['autoaxis_x'] == 'on':
                    self.zoom_fit_x()
                else:
                    if not self._panning_x and self.settings['autoaxis_x'] == 'pan':
                        self.autopan_x()

                # recalculate plots based on new X limits
                bbox = self.viewer.get_pan_bbox()
                x_lo, x_hi = bbox[0][0], bbox[2][0]

                for plot_src in self.plots.values():
                    plot_src.calc_points(self.viewer, x_lo, x_hi)

                # Y limits may be adjusted by autoscaling Y axis
                if not self._scaling_y and self.settings['autoaxis_y'] != 'off':
                    self.zoom_fit_y()

                # finally, update plot elements to match adjustments
                self.update_elements()

        finally:
            self._adjusting = False

    def zoom_limit_y(self, y_lo, y_hi):
        """Scale the plot in the Y direction so that `y_lo` and `y_hi`
        are both visible.

        Parameters
        ----------
        y_lo : `float` or `int`
            Low point of Y plot to show

        y_hi : `float` or `int`
            High point of Y plot to show
        """
        self._scaling_y = True
        try:
            # get current viewer limits for X
            vl = self.viewer.get_limits()
            x_lo, x_hi = vl[0][0], vl[1][0]

            y_lo, y_hi = self.tweak_limits(y_lo, y_hi)
            self.viewer.set_limits([(x_lo, y_lo), (x_hi, y_hi)])
            self.viewer.zoom_fit(axis='y')

        finally:
            self._scaling_y = False

    def zoom_fit_y(self):
        """Scale the plot in the Y axis so that the plot takes up as
        much room vertically as shown in the visible area.
        """
        # establish range of current Y plot
        mode = self.settings['autoaxis_y']
        if mode == 'on':
            limits = self.get_limits('data')
        else:
            limits = self.get_limits('plot')

        y_lo, y_hi = limits[0][1], limits[1][1]
        self.zoom_limit_y(y_lo, y_hi)

    def zoom_fit_y_full(self):
        """Scale the plot in the Y axis so that the plot takes up as
        much room vertically as there is in the range of Y data.
        """
        # establish full range of Y
        limits = self.get_limits('data')
        y_lo, y_hi = limits[0][1], limits[1][1]

        self.zoom_limit_y(y_lo, y_hi)

    def zoom_limit_x(self, x_lo, x_hi):
        """Scale the plot in the X direction so that `x_lo` and `x_hi`
        are both visible.

        Parameters
        ----------
        x_lo : `float` or `int`
            Low point of X plot to show

        x_hi : `float` or `int`
            High point of X plot to show
        """
        self._scaling_x = True
        try:
            # get current viewer limits for Y
            vl = self.viewer.get_limits()
            y_lo, y_hi = vl[0][1], vl[1][1]

            self.viewer.set_limits([(x_lo, y_lo), (x_hi, y_hi)])
            self.viewer.zoom_fit(axis='x')

        finally:
            self._scaling_x = False

    def zoom_fit_x(self):
        """Scale the plot in the X axis so that the plot takes up as
        much room horizontally as there is in the range of X data.
        """
        # establish full range of X
        limits = self.get_limits('data')
        x_lo, x_hi = limits[0][0], limits[1][0]

        self.zoom_limit_x(x_lo, x_hi)

    def autopan_x(self):
        """Pan the plot so that the most recent data is visible at
        the right side of the plot. i.e. "keep up with the plot"
        """
        self._panning_x = True
        try:
            wd, ht = self.viewer.get_window_size()

            pl = self.get_limits('data')
            x_lo, x_hi = pl[0][0], pl[1][0]
            pan_x, pan_y = self.viewer.get_pan()[:2]
            _, cy = self.viewer.get_canvas_xy(pan_x, pan_y)
            cx = wd

            self.viewer.position_at_canvas_xy((x_hi, pan_y), (cx, cy))

        finally:
            self._panning_x = False

    def _autoaxis_x_cb(self, setting, autoaxis_x):
        """Called when the user toggles the 'autoaxis X' feature on or off.
        """
        if autoaxis_x == 'on':
            msg = "Autoaxis X ON"
        elif autoaxis_x == 'pan':
            msg = "Autoaxis X PAN"
        else:
            msg = "Autoaxis X OFF"
        self.viewer.onscreen_message(msg, delay=1.0)
        self.adjust_view()

    def _autoaxis_y_cb(self, setting, autoaxis_y):
        """Called when the user toggles the 'autoaxis Y' feature on or off.
        """
        if autoaxis_y == 'on':
            msg = "Autoaxis Y ON"
        elif autoaxis_y == 'vis':
            msg = "Autoaxis Y VIS"
        else:
            msg = "Autoaxis Y OFF"
        self.viewer.onscreen_message(msg, delay=1.0)
        self.adjust_view()

    def _autopan_x_cb(self, setting, autopan_x):
        """Called when the user toggles the 'autopan X' feature on or off.
        """
        msg = "Autoaxis X PAN" if autopan_x == 'pan' else "Autoaxis X OFF"
        self.viewer.onscreen_message(msg, delay=1.0)
        self.adjust_view()

    def adjust_resize_cb(self):
        """Callback that is called when the viewer window is resized."""
        with self.viewer.suppress_redraw:
            self.update_resize()
            self.adjust_view()

    def pan_cb(self, plot, pan_pos):
        """Called when the user pans the plot viewer."""
        if self._panning_x:
            return
        t = pan_pos[0]
        _t, y = self.viewer.get_pan()[:2]
        self.viewer.set_pan(t, y)
        self.adjust_view()

    def plot_zoom_x_cb(self, plot, direction):
        """Default callback called when the user zooms the plot in X."""
        scale_x, scale_y = self.viewer.get_scale_xy()
        zoomrate = self.viewer.get_zoomrate()

        with self.viewer.suppress_redraw:
            # zoom X axis
            if direction == 'in':
                scale_x *= zoomrate
            elif direction == 'out':
                scale_x *= (1.0 / zoomrate)

            self.viewer.scale_to(scale_x, scale_y)
            # turn off X autoscaling, since user overrode it
            if self.settings['autoaxis_x'] == 'on':
                self.settings['autoaxis_x'] = 'off'

            self.adjust_view()

        return True

    def plot_zoom_y_cb(self, plot, direction):
        """Default callback called when the user zooms the plot in Y."""
        scale_x, scale_y = self.viewer.get_scale_xy()
        zoomrate = self.viewer.get_zoomrate()

        with self.viewer.suppress_redraw:
            # zoom Y axis
            if direction == 'in':
                scale_y *= 1.1
            elif direction == 'out':
                scale_y *= 0.9

            self.viewer.scale_to(scale_x, scale_y)
            # turn off Y autoscaling, since user overrode it
            if self.settings['autoaxis_y'] != 'off':
                self.settings['autoaxis_y'] = 'off'

            self.adjust_view()

        return True

    def _pan_cb(self, settings, pan_pos):
        """Callback invoked by the viewer when a pan happens pans in our
        particular plot viewer. We use this to adjust the view if the pan is
        not a byproduct of an internal adjustment.
        """
        # check to avoid circular callback situation
        if not (self._panning_x or self._scaling_x or self._scaling_y or
                self._adjusting):
            self.adjust_view()

    def update_horz_scrollbar_cb(self, setting, autoaxis_x, sw):
        h = 'off' if autoaxis_x != 'off' else 'on'
        d = sw.get_scroll_bars_status()
        v = d['vertical']
        sw.scroll_bars(horizontal=h, vertical=v)

    def update_vert_scrollbar_cb(self, setting, autoaxis_y, sw):
        d = sw.get_scroll_bars_status()
        h = d['horizontal']
        v = 'off' if autoaxis_y != 'off' else 'on'
        sw.scroll_bars(horizontal=h, vertical=v)

    def make_tt(self, canvas, lines, pt, fontsize=10):
        # create the canvas object representing the tooltip
        # a Rectangle with some Text objects inside
        rect = canvas.dc.Rectangle(0, 0, 0, 0, color='black', fill=True,
                                   fillcolor='lightyellow', coord='window')

        l = [rect]
        for line in lines:
            text = canvas.dc.Text(0, 0, text=line, color='black',
                                  font='sans condensed', fontsize=fontsize,
                                  coord='window')
            l.append(text)

        tt_obj = canvas.dc.CompoundObject(*l)
        return tt_obj

    def update_tt(self, tt_obj, lines, pt, fontsize=10):
        # Determine pop-up position on canvas, offset a little from cursor.
        x, y = pt[:2]
        x, y = x + 15, y + 10
        mxwd = 0

        # quick check to make sure TT doesn't go out of plot area in Y
        text = tt_obj.objects[1]
        text.text = lines[0]
        txt_wd, txt_ht = self.viewer.renderer.get_dimensions(text)
        y_end = y + len(lines) * (txt_ht + 1)
        y_diff = y_end - self.llur[3]
        if y_diff > 0:
            y -= y_diff

        # assign background rect coords
        rect = tt_obj.objects[0]
        rect.x1, rect.y1 = x, y

        # assign text coords
        a, b = x + 2, y
        for i, text in enumerate(tt_obj.objects[1:]):
            text.x = a
            text.text = lines[i]
            txt_wd, txt_ht = self.viewer.renderer.get_dimensions(text)
            b += txt_ht + 1
            text.y = b
            mxwd = max(mxwd, txt_wd)
        rect.x2 = x + mxwd + 4
        rect.y2 = b + 4

    def show_tt_cb(self, plot_bg, canvas, event, pt, state, fontsize=10):
        """Internal callback routine to pop up a tooltip-like reading of
        X/Y values at the cursor.
        """
        # determine plot coordinates of cursor
        # double conversion needed because plot transform is different
        # from default data transform
        win_pt = self.viewer.tform['data_to_window'].to_(pt)
        plot_x, plot_y = self.viewer.tform['data_to_plot'].from_(win_pt)

        # format these values according to the axes formatters
        x_axis = self.get_plot_decor('axis_x')
        x_text = x_axis.format_value(plot_x)
        y_axis = self.get_plot_decor('axis_y')
        y_text = y_axis.format_value(plot_y)
        text = "X: {0:}\nY: {1:}".format(x_text, y_text)
        lines = text.split('\n')

        tag = '_$tooltip'
        if state == 'down':
            # pressed down, create and pop up tooltip with X/Y value
            # at point
            try:
                canvas.delete_object_by_tag(tag, redraw=False)
            except KeyError:
                pass
            tt = self.make_tt(canvas, lines, win_pt)
            canvas.add(tt, tag=tag, redraw=False)
            self.update_tt(tt, lines, win_pt, fontsize=fontsize)

        elif state == 'move':
            # cursor moved while pressed down; update the tooltip value
            tt = canvas.get_object_by_tag(tag)
            self.update_tt(tt, lines, win_pt)

        else:
            # up, or anything else, remove the tooltip
            try:
                canvas.delete_object_by_tag(tag, redraw=False)
            except KeyError:
                pass

        canvas.update_canvas()
        return True



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga import trcalc
from ginga.gw import Widgets

from .base import Stage, StageAction


class FlipSwap(Stage):

    _stagename = 'flip-swap'

    def __init__(self):
        super().__init__()

        self._flip_x = False
        self._flip_y = False
        self._swap_xy = False
        self.viewer = None

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')

        # TRANSFORM OPTIONS
        fr = Widgets.Frame("Flip / Swap")

        captions = (('Transform:', 'label', 'hbox1', 'hbox'),
                    ('_sp1', 'spacer', 'hbox2', 'hbox'),
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')

        for wname, name, tf in (('flip_x', "Flip X", self._flip_x),
                                ('flip_y', "Flip Y", self._flip_y),
                                ('swap_xy', "Swap XY", self._swap_xy)):
            btn = Widgets.CheckBox(name)
            b[wname] = btn
            btn.set_state(tf)
            btn.add_callback('activated', self.set_transforms_cb)
            b.hbox1.add_widget(btn, stretch=0)

        b.flip_x.set_tooltip("Flip the image around the X axis")
        b.flip_y.set_tooltip("Flip the image around the Y axis")
        b.swap_xy.set_tooltip("Swap the X and Y axes in the image")

        b.copy_from_viewer = Widgets.Button("Copy from viewer")
        b.copy_from_viewer.set_tooltip("Copy flip/swap setting from viewer")
        b.copy_from_viewer.add_callback('activated', self.copy_from_viewer_cb)
        b.hbox2.add_widget(b.copy_from_viewer, stretch=0)
        b.hbox2.add_widget(Widgets.Label(''), stretch=1)

        self.w.update(b)
        fr.set_widget(w)

        container.set_widget(fr)

    @property
    def flip_x(self):
        return self._flip_x

    @flip_x.setter
    def flip_x(self, tf):
        self._flip_x = tf
        if self.gui_up:
            self.w.flip_x.set_state(tf)

    @property
    def flip_y(self):
        return self._flip_y

    @flip_y.setter
    def flip_y(self, tf):
        self._flip_y = tf
        if self.gui_up:
            self.w.flip_y.set_state(tf)

    @property
    def swap_xy(self):
        return self._swap_xy

    @swap_xy.setter
    def swap_xy(self, tf):
        self._swap_xy = tf
        if self.gui_up:
            self.w.swap_xy.set_state(tf)

    def _get_state(self):
        return dict(flip_x=self._flip_x, flip_y=self._flip_y,
                    swap_xy=self._swap_xy)

    def set_transforms_cb(self, *args):
        old = self._get_state()
        self._flip_x = self.w.flip_x.get_state()
        self._flip_y = self.w.flip_y.get_state()
        self._swap_xy = self.w.swap_xy.get_state()
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="flip / swap"))
        self.pipeline.run_from(self)

    def copy_from_viewer_cb(self, widget):
        old = self._get_state()
        self.flip_x, self.flip_y, self.swap_xy = self.viewer.get_transforms()
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="flip / swap"))
        self.pipeline.run_from(self)

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        res_np = trcalc.transform(data, flip_x=self.flip_x, flip_y=self.flip_y,
                                  swap_xy=self.swap_xy)

        self.pipeline.send(res_np=res_np)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.flip_x = d['flip_x']
        self.flip_y = d['flip_y']
        self.swap_xy = d['swap_xy']



#
# PilHelp.py -- help classes for the PIL drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import numpy as np
from PIL import Image, ImageFont, ImageDraw

from ginga import colors
from ginga.fonts import font_asst


def get_cached_font(fontname, fontsize):
    key = ('pil', fontname, fontsize)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        # see if we can build the font
        info = font_asst.get_font_info(fontname, subst_ok=True)

        font = ImageFont.truetype(info.font_path, fontsize)
        font_asst.add_cache(key, font)

        return font


def load_font(font_name, font_file):
    if not font_asst.have_font(font_name):
        font_asst.add_font(font_file, font_name=font_name)
    return font_name


def text_size(text, font):
    f = get_cached_font(font.fontname, font.fontsize)
    if hasattr(f, 'getbbox'):
        # PIL v10.0
        l, t, r, b = f.getbbox(text)
        wd_px, ht_px = int(abs(round(r - l))), int(abs(round(b - t)))
    else:
        wd_px, ht_px = f.getsize(text)
    return wd_px, ht_px


def text_to_array(text, font, rot_deg=0.0):
    wd, ht = text_size(text, font)
    f = get_cached_font(font.fontname, font.fontsize)
    color = get_color(font.color)
    i = Image.new('RGBA', (wd, ht))
    d = ImageDraw.Draw(i, 'RGBA')
    d.text((0, 0), text, font=f, fill=color)
    i.rotate(rot_deg, expand=1)
    arr8 = np.frombuffer(i.tobytes(), dtype=np.uint8)
    arr8 = arr8.reshape((ht, wd, 4))
    return arr8


def get_color(color, alpha=1.0):
    if color is not None:
        r, g, b = colors.resolve_color(color)
    else:
        r, g, b = 1.0, 1.0, 1.0

    return (int(r * 255), int(g * 255), int(b * 255), int(alpha * 255))


class PilContext(object):

    def __init__(self, surface):
        self.set_canvas(surface)

    def set_canvas(self, surface):
        self.surface = surface
        self.ctx = ImageDraw.Draw(surface, 'RGBA')

    def _cvt_points(self, points):
        # PIL seems to have trouble with numpy arrays as sequences
        # of points, so just convert to a list

        # TODO: this doesn't work--would be more efficient if it did
        # if isinstance(points, np.ndarray):
        #     return points.tolist()
        # return points
        return [(p[0], p[1]) for p in points]

    def _get_font(self, font):
        font = get_cached_font(font.fontname, font.fontsize)
        return font

    def text_extents(self, text, font):
        return text_size(text, font)

    def image(self, pt, rgb_arr):
        p_image = Image.fromarray(rgb_arr)

        self.surface.paste(p_image)

    def text(self, pt, text, font, line, fill):
        x, y = pt
        kwargs = dict()
        if font is not None:
            kwargs['font'] = font.render.font
        if fill is not None:
            kwargs['fill'] = fill.render.color
        if line is not None:
            kwargs['stroke_width'] = int(line.linewidth)
            kwargs['stroke_fill'] = line.render.color
        self.ctx.text((x, y), text, **kwargs)

    def line(self, pt1, pt2, line):
        if line is not None:
            x1, y1 = int(np.round(pt1[0])), int(np.round(pt1[1]))
            x2, y2 = int(np.round(pt2[0])), int(np.round(pt2[1]))
            self.ctx.line(((x1, y1), (x2, y2)), fill=line.render.color,
                          width=line.linewidth)

    def circle(self, pt, radius, line, fill):
        x, y = pt
        radius = int(radius)
        kwargs = dict()
        if fill is not None:
            kwargs['fill'] = fill.render.color
        if line is not None:
            kwargs['width'] = line.linewidth
            kwargs['outline'] = line.render.color
        self.ctx.ellipse(((x - radius, y - radius), (x + radius, y + radius)),
                         **kwargs)

    def polygon(self, points, line, fill):
        points = self._cvt_points(points)

        kwargs = dict()
        if fill is not None:
            kwargs['fill'] = fill.render.color
        if line is not None:
            kwargs['width'] = line.linewidth
            kwargs['outline'] = line.render.color
        self.ctx.polygon(points, **kwargs)

    def path(self, points, line):
        points = self._cvt_points(points)

        p0 = points[0]
        for pt in points[1:]:
            self.line(p0, pt, line)
            p0 = pt

#END



#
# CanvasRenderCairo.py -- for rendering into a ImageViewCairo widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import sys

import numpy as np
import cairo

from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa

from . import CairoHelp


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        self.ctx = cairo.Context(surface)
        # set up antialiasing
        fo = cairo.FontOptions()
        fo.set_antialias(cairo.ANTIALIAS_DEFAULT)
        self.ctx.set_font_options(fo)

        # special scaling for Cairo text drawing to normalize it relative
        # to other backends
        self._font_scale_factor = 1.75

    def setup_line(self, line):
        r, g, b, a = line._color_4tup
        self.ctx.set_source_rgba(r, g, b, a)
        self.ctx.set_line_width(line.linewidth)
        if line.linestyle == 'dash':
            self.ctx.set_dash([3.0, 4.0, 6.0, 4.0], 5.0)

    def setup_fill(self, fill):
        r, g, b, a = fill._color_4tup
        self.ctx.set_source_rgba(r, g, b, a)

    def setup_font(self, font):
        self.ctx.select_font_face(font.fontname)
        self.ctx.set_font_size(font.fontsize)

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        wd, ht = CairoHelp.text_size(text, font)
        return wd, ht

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, rgb_arr, whence, order='RGBA'):
        # no-op for this renderer
        pass

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, line=None,
                  fill=None):
        self.ctx.save()
        self.ctx.translate(cx, cy)
        self.ctx.move_to(0, 0)
        self.ctx.rotate(-np.radians(rot_deg))
        if font is not None:
            self.setup_font(font)
        if line is not None:
            self.setup_line(line)
            self.ctx.text_path(text)
            self.ctx.stroke_preserve()
            if fill is not None:
                self.setup_fill(fill)
                self.ctx.fill()
        else:
            # NOTE: show_text() is more efficient if we have a lot of text,
            # according to Cairo docs, so use it instead if we are not
            # stroking a path outline
            if fill is not None:
                self.setup_fill(fill)
                self.ctx.show_text(text)

        self.ctx.restore()
        self.ctx.new_path()

    def draw_polygon(self, cpoints, line=None, fill=None):
        (cx0, cy0) = cpoints[-1][:2]
        self.ctx.move_to(cx0, cy0)
        for cpt in cpoints:
            cx, cy = cpt[:2]
            self.ctx.line_to(cx, cy)
        self.ctx.close_path()
        if line is not None:
            self.setup_line(line)
            self.ctx.stroke_preserve()
        if fill is not None:
            self.setup_fill(fill)
            self.ctx.fill()
        self.ctx.new_path()

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        self.ctx.arc(cx, cy, cradius, 0, 2 * np.pi)
        if line is not None:
            self.setup_line(line)
            self.ctx.stroke_preserve()
        if fill is not None:
            self.setup_fill(fill)
            self.ctx.fill()
        self.ctx.new_path()

    def draw_bezier_curve(self, cp, line=None):
        self.ctx.move_to(cp[0][0], cp[0][1])
        self.ctx.curve_to(cp[1][0], cp[1][1], cp[2][0], cp[2][1], cp[3][0], cp[3][1])

        if line is not None:
            self.setup_line(line)
            self.ctx.stroke()
        self.ctx.new_path()

    def draw_ellipse_bezier(self, cp, line=None, fill=None):
        # draw 4 bezier curves to make the ellipse
        self.ctx.move_to(cp[0][0], cp[0][1])
        self.ctx.curve_to(cp[1][0], cp[1][1], cp[2][0], cp[2][1], cp[3][0], cp[3][1])
        self.ctx.curve_to(cp[4][0], cp[4][1], cp[5][0], cp[5][1], cp[6][0], cp[6][1])
        self.ctx.curve_to(cp[7][0], cp[7][1], cp[8][0], cp[8][1], cp[9][0], cp[9][1])
        self.ctx.curve_to(cp[10][0], cp[10][1], cp[11][0], cp[11][1], cp[12][0], cp[12][1])
        if line is not None:
            self.setup_line(line)
            self.ctx.stroke_preserve()

        if fill is not None:
            self.setup_fill(fill)
            self.ctx.fill()
        self.ctx.new_path()

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        self.ctx.set_line_cap(cairo.LINE_CAP_ROUND)
        self.ctx.move_to(cx1, cy1)
        self.ctx.line_to(cx2, cy2)
        if line is not None:
            self.setup_line(line)
            self.ctx.stroke()
        self.ctx.new_path()

    def draw_path(self, cpoints, line=None):
        (cx0, cy0) = cpoints[0][:2]
        self.ctx.move_to(cx0, cy0)
        for cpt in cpoints[1:]:
            cx, cy = cpt[:2]
            self.ctx.line_to(cx, cy)
        if line is not None:
            self.setup_line(line)
            self.ctx.stroke()
        self.ctx.new_path()


class CanvasRenderer(render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)

        self.kind = 'cairo'
        if sys.byteorder == 'little':
            self.rgb_order = 'BGRA'
        else:
            self.rgb_order = 'ARGB'
        self.surface = None
        self.surface_arr = None

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        width, height = dims[:2]
        self.logger.debug("renderer reconfigured to %dx%d" % (
            width, height))

        # create cairo surface the size of the window
        #surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        depth = len(self.rgb_order)
        self.surface_arr = np.zeros((height, width, depth), dtype=np.uint8)

        stride = cairo.ImageSurface.format_stride_for_width(cairo.FORMAT_ARGB32,
                                                            width)
        surface = cairo.ImageSurface.create_for_data(self.surface_arr,
                                                     cairo.FORMAT_ARGB32,
                                                     width, height, stride)
        self.surface = surface

        # fill surface with background color;
        # this reduces unwanted garbage in the resizing window
        ctx = cairo.Context(self.surface)

        # fill surface with background color
        ctx.rectangle(0, 0, width, height)
        r, g, b = self.viewer.get_bg()
        ctx.set_source_rgba(r, g, b)
        ctx.fill()

        super(CanvasRenderer, self).resize(dims)

    def render_image(self, arr, order, win_coord):
        """Render the image represented by (data) at (win_coord)
        in the pixel space.
        *** internal method-- do not use ***
        """
        self.logger.debug("redraw surface")
        if self.surface is None:
            return

        dst_x, dst_y = win_coord[:2]

        daht, dawd, depth = arr.shape
        self.logger.debug("arr shape is %dx%dx%d" % (dawd, daht, depth))

        ctx = cairo.Context(self.surface)
        # TODO: is it really necessary to hang on to this context?
        self.ctx = ctx

        # fill surface with background color
        imgwin_wd, imgwin_ht = self.viewer.get_window_size()
        ctx.rectangle(0, 0, imgwin_wd, imgwin_ht)
        r, g, b = self.viewer.get_bg()
        ctx.set_source_rgba(r, g, b)
        ctx.fill()

        stride = cairo.ImageSurface.format_stride_for_width(cairo.FORMAT_ARGB32,
                                                            dawd)
        img_surface = cairo.ImageSurface.create_for_data(arr,
                                                         cairo.FORMAT_ARGB32,
                                                         dawd, daht, stride)

        ctx.set_source_surface(img_surface, dst_x, dst_y)
        ctx.set_operator(cairo.OPERATOR_SOURCE)

        ctx.mask_surface(img_surface, dst_x, dst_y)
        ctx.fill()

    def get_surface_as_array(self, order=None):
        if self.surface_arr is None:
            raise render.RenderError("No cairo surface defined")

        # adjust according to viewer's needed order
        src_order = self.get_rgb_order()
        return self.reorder(order, self.surface_arr, src_order=src_order)

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        font = cr.get_font_from_shape(shape)
        return cr.text_extents(shape.text, font=font)

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.surface)
        cr.setup_font(font)
        return cr.text_extents(text)

# END



#
# VecHelp.py -- help classes for vector drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

IMAGE = 0
LINE = 1
CIRCLE = 2
BEZIER = 3
ELLIPSE_BEZIER = 4
POLYGON = 5
PATH = 6
TEXT = 7



"""
Test program for trying widgets in the different wrapped toolkits supported
by Ginga.

Usage:
  $ python widgets.py <toolkit-name> [logging options]

Examples:
  $ python widgets.py -t qt5
  $ python widgets.py -t gtk3
  $ python widgets.py -t pg
"""
import sys
import os

from ginga.misc import log
from ginga.util.paths import icondir  # noqa

top = None


def quit(*args):
    if top is not None:
        top.delete()
    sys.exit()


def popup_dialog(parent):
    from ginga.gw import Widgets
    dia = Widgets.Dialog(title="Dialog Title",
                         buttons=[('ok', 0), ('cancel', 1)],
                         parent=parent, modal=True)
    cntr = dia.get_content_area()
    cntr.add_widget(Widgets.Label("My Dialog Content"))
    parent.add_dialog(dia)
    dia.show()


def button_test(w):
    if w.get_text().lower() == 'foo':
        w.set_text('bar')
    else:
        w.set_text('foo')


def setting_limits(textentry, spinbox):
    limits = textentry.get_text().split(",")
    spinbox.set_limits(float(limits[0]), float(limits[1]), float(limits[2]))


def grid_helper(textentry, gridbox):
    from ginga.gw import Widgets
    indices = textentry.get_text().split(",")
    gridbox.add_widget(Widgets.Label("Added widget"), int(indices[0]), int(indices[1]))


def show_example(cbox, top, logger):
    from ginga.gw import Widgets
    wname = cbox.get_text()
    vbox = Widgets.VBox()
    vbox.set_border_width(2)
    vbox.set_spacing(1)

    if wname == 'label':
        _vbox = Widgets.VBox()
        label = Widgets.Label("Hello World label")
        label.set_color('white', 'blue')
        label.set_font("Times New Roman")
        _vbox.add_widget(label, stretch=1)
        # Allow user to select different font
        fontbox = Widgets.ComboBox()
        _vbox.add_widget(fontbox, stretch=0)
        for name in ["Times New Roman", "Arial", "Courier"]:
            fontbox.append_text(name)
        fontbox.add_callback('activated', lambda r, val: label.set_font(r.get_text()))
        fontbox.add_callback('activated',
                             lambda w, val: logger.info("chose '{}'".format(w.get_text())))
        # ...FOR HALIGN...
        alignbox = Widgets.ComboBox()
        _vbox.add_widget(alignbox, stretch=0)
        for name in ["Left", "Center", "Right"]:
            alignbox.append_text(name)
        alignbox.add_callback('activated', lambda r, val: label.set_halign(r.get_text()))
        alignbox.add_callback('activated',
                              lambda w, val: logger.info("chose '{}'".format(w.get_text())))
        label.set_halign('Left')
        vbox.add_widget(_vbox)

    elif wname == 'button':
        w = Widgets.Button("Press me")
        w.add_callback('activated', lambda w: logger.info("button was clicked"))
        w.add_callback('activated', button_test)
        #w.add_callback('activated', lambda w: popup_dialog(top))
        vbox.add_widget(w, stretch=1)

    elif wname == 'textentry':
        _vbox = Widgets.VBox()
        # Read only text entry
        w2 = Widgets.TextEntry("Read only", editable=False)
        _vbox.add_widget(w2, stretch=1)
        # Regular text entry
        w = Widgets.TextEntry("Hello, World!")
        w.set_font("Times New Roman")
        w.add_callback('activated',
                       lambda w: logger.info("said '{}'".format(w.get_text())))
        w.add_callback('activated', lambda r: w.set_text(r.get_text()))
        _vbox.add_widget(w, stretch=1)
        # Allow user to select different font
        fontbox = Widgets.ComboBox()
        _vbox.add_widget(fontbox, stretch=0)
        for name in ["Times New Roman", "Helvetica", "Courier"]:
            fontbox.append_text(name)
        fontbox.add_callback('activated', lambda r, val: w.set_font(r.get_text()))
        # Set length
        hbox = Widgets.HBox()
        label = Widgets.Label("Set length: ")
        enter_limit = Widgets.TextEntrySet()
        enter_limit.add_callback('activated', lambda r: w.set_length(int(enter_limit.get_text())))
        hbox.add_widget(label)
        hbox.add_widget(enter_limit)
        _vbox.add_widget(hbox, stretch=0)

        vbox.add_widget(_vbox, stretch=1)

    elif wname == 'textentryset':
        _vbox = Widgets.VBox()
        # Read only text entry
        w2 = Widgets.TextEntrySet("Read only", editable=False)
        _vbox.add_widget(w2, stretch=1)
        # Editable text entry
        w = Widgets.TextEntrySet()
        w.set_text("This text entry fits 35 characters!")
        w.set_length(35)
        w.set_font("Georgia")
        w.add_callback('activated',
                       lambda w: logger.info("said '{}'".format(w.get_text())))
        _vbox.add_widget(w, stretch=1)
        vbox.add_widget(_vbox, stretch=1)

    elif wname == 'textarea':
        _vbox = Widgets.VBox()
        # Read only text area
        w2 = Widgets.TextArea(editable=False)
        w2.set_text("READ ONLY")
        _vbox.add_widget(w2, stretch=1)
        # Editable text area
        w = Widgets.TextArea(editable=True)
        w.set_text("Hello, World!")
        w.set_wrap(False)
        _vbox.add_widget(w, stretch=1)
        # Set wrap
        b2 = Widgets.ToggleButton("Toggle wrap on/off")
        b2.add_callback('activated', lambda r, val: w.set_wrap(b2.get_state()))
        _vbox.add_widget(b2)
        # Clear text
        b = Widgets.Button("Clear")
        b.add_callback('activated', lambda r: w.clear())
        _vbox.add_widget(b)
        # Select font
        fontbox = Widgets.ComboBox()
        w.set_font("Arial")
        for name in ["Times New Roman", "Helvetica", "Courier"]:
            fontbox.append_text(name)
        fontbox.add_callback('activated', lambda r, val: w.set_font(r.get_text()))
        _vbox.add_widget(fontbox, stretch=0)

        vbox.add_widget(_vbox)

    elif wname == 'checkbox':
        w = Widgets.CheckBox("Check me")
        w.set_state(True)
        w.add_callback('activated',
                       lambda w, val: logger.info("value changed to '{}'".format(w.get_state())))
        vbox.add_widget(w, stretch=1)

    elif wname == 'dial':
        w = Widgets.Dial()
        w.set_limits(-20, 20, incr_value=1)
        w.add_callback('value-changed',
                       lambda w, val: logger.info("value changed to '{}'".format(w.get_value())))
        vbox.add_widget(w)

    elif wname == 'togglebutton':
        w = Widgets.ToggleButton("Toggle me")
        w.set_state(True)
        w.add_callback('activated',
                       lambda w, val: logger.info("value changed to '{}'".format(w.get_state())))
        vbox.add_widget(w, stretch=1)

    elif wname == 'radiobutton':
        w = Widgets.RadioButton("Option 1")
        w.add_callback('activated',
                       lambda w, val: logger.info("chose option 1"))
        vbox.add_widget(w)
        w2 = Widgets.RadioButton("Option 2", group=w)
        w2.add_callback('activated',
                        lambda w, val: logger.info("chose option 2"))
        vbox.add_widget(w2)
        w3 = Widgets.RadioButton("Option 3", group=w)
        w3.set_state(True)
        w3.add_callback('activated',
                        lambda w, val: logger.info("chose option 3"))
        vbox.add_widget(w3)

    elif wname == 'combobox':
        w = Widgets.ComboBox()
        for name in ["Larry", "Curly", "Moe"]:
            w.append_text(name)
        w.add_callback('activated',
                       lambda w, val: logger.info("chose '{}'".format(w.get_text())))
        vbox.add_widget(w, stretch=0)
        # Test insert_alpha
        hbox1 = Widgets.HBox()
        label1 = Widgets.Label("Add Choice: ")
        insert = Widgets.TextEntrySet()
        insert.add_callback('activated', lambda r: w.insert_alpha(r.get_text()))
        hbox1.add_widget(label1, stretch=0)
        hbox1.add_widget(insert, stretch=0)
        vbox.add_widget(hbox1, stretch=0)
        # Test delete_alpha
        hbox2 = Widgets.HBox()
        label2 = Widgets.Label("Delete choice: ")
        delete = Widgets.TextEntrySet()
        delete.add_callback('activated', lambda r: w.delete_alpha(r.get_text()))
        hbox2.add_widget(label2, stretch=0)
        hbox2.add_widget(delete, stretch=0)
        vbox.add_widget(hbox2, stretch=0)
        # Test clear method
        clearbutton = Widgets.Button("Clear choices")
        clearbutton.add_callback('activated', lambda r: w.clear())
        vbox.add_widget(clearbutton, stretch=0)

    elif wname == 'spinbox':
        w = Widgets.SpinBox(dtype=float)
        w.set_limits(-10, 10, incr_value=0.1)
        w.set_value(5.5)
        # Displaying changing value
        w.add_callback('value-changed',
                       lambda w, val: logger.info("chose {}".format(val)))
        vbox.add_widget(w)
        value_label = Widgets.Label("Value: {}".format(w.get_value()))
        w.add_callback('value-changed', lambda r, val: value_label.set_text("Value: {}".format(val)))
        vbox.add_widget(value_label, stretch=1)
        # Test set_limits dynamically
        limits_label = Widgets.Label("Set limits (min, max, incr): ")
        change_limits = Widgets.TextEntrySet()
        hbox = Widgets.HBox()
        hbox.add_widget(limits_label, stretch=1)
        hbox.add_widget(change_limits, stretch=1)
        change_limits.add_callback('activated', lambda r: setting_limits(r, w))
        vbox.add_widget(hbox, stretch=1)

    elif wname == 'slider':
        _vbox = Widgets.VBox()
        w = Widgets.Slider(orientation='horizontal')
        w.set_limits(0, 5, incr_value=1)
        w.set_value(4)
        w.set_tracking(True)
        w.add_callback('value-changed',
                       lambda w, val: logger.info("chose {}".format(val)))
        _vbox.add_widget(w)
        # Display current value
        value_label = Widgets.Label("Value: {}".format(w.get_value()))
        w.add_callback('value-changed', lambda r, val: value_label.set_text("Value: {}".format(val)))
        _vbox.add_widget(value_label)
        # Test set_tracking
        b = Widgets.CheckBox("Tracking on/off")
        b.add_callback('activated', lambda wd, tf: w.set_tracking(tf))
        _vbox.add_widget(b)
        # Test set_value dynamically
        hbox = Widgets.HBox()
        set_label = Widgets.Label("Set value: ")
        change_val = Widgets.TextEntrySet()
        change_val.add_callback('activated', lambda r: w.set_value(int(r.get_text())))
        hbox.add_widget(set_label)
        hbox.add_widget(change_val, stretch=1)
        _vbox.add_widget(hbox, stretch=1)
        # Test set_limits dynamically
        hbox2 = Widgets.HBox()
        limits_label = Widgets.Label("Set limits (min, max, incr): ")
        change_limits = Widgets.TextEntrySet()
        hbox2.add_widget(limits_label)
        hbox2.add_widget(change_limits)
        change_limits.add_callback('activated', lambda r: setting_limits(r, w))
        _vbox.add_widget(hbox2, stretch=1)
        vbox.add_widget(_vbox)

    elif wname == 'scrollbar':
        _vbox = Widgets.VBox()
        w = Widgets.ScrollBar(orientation='horizontal')
        _vbox.add_widget(w, stretch=0)
        _ent = Widgets.TextEntrySet()
        _vbox.add_widget(_ent, stretch=0)
        _ent.add_callback('activated', lambda r: w.set_value(float(r.get_text())))
        w.set_value(0.40)
        w.add_callback('activated', lambda w, val: logger.info("value is %.2f" % val))
        vbox.add_widget(_vbox)

    elif wname == 'progressbar':
        _vbox = Widgets.VBox()
        w = Widgets.ProgressBar()
        _vbox.add_widget(w, stretch=0)
        _ent = Widgets.TextEntrySet()
        _vbox.add_widget(_ent, stretch=0)
        _ent.add_callback('activated', lambda r: w.set_value(float(r.get_text())))
        w.set_value(0.6)
        vbox.add_widget(_vbox)

    elif wname == 'statusbar':
        w = Widgets.StatusBar()
        w.set_message("Hello, World! is my status")
        vbox.add_widget(w)

    elif wname == 'image':
        w = Widgets.Image()
        w.load_file(os.path.join(icondir, 'ginga-512x512.png'))
        vbox.add_widget(w)

    elif wname == 'treeview':
        w = Widgets.TreeView(selection='single', sortable=True,
                             use_alt_row_color=True)
        columns = [("Meal", 'meal'), ("Critic 1", 'review1'),
                   ("Critic 2", 'review2'), ("Critic 3", 'review3')]
        w.setup_table(columns, 1, 'meal')
        tree = dict(Breakfast=dict(meal='Breakfast', review1="Delish!",
                                   review2="Ugh!", review3="Meh"),
                    Lunch=dict(meal='Lunch', review1="Gross!",
                               review2="Interesting...", review3="Meh"),
                    Supper=dict(meal='Supper', review1="Meh",
                                review2="Meh", review3="Jolly good!"))
        w.set_tree(tree)
        vbox.add_widget(w, stretch=1)

    elif wname == 'webview':
        w = Widgets.WebView()
        w.load_url("http://www.google.com/")
        vbox.add_widget(w)

    elif wname == 'frame':
        w = Widgets.Frame(title="Frame Title")
        slider = Widgets.Slider(orientation='horizontal')
        slider.set_limits(-10, 10, incr_value=1)
        w.set_widget(slider)
        vbox.add_widget(w)
        # Test set_widget
        button = Widgets.Button("Press me to set to another widget!")
        button.add_callback('activated', lambda r: w.set_widget(Widgets.TextArea(editable=True)))
        # Test set_text to change header inside frame widget
        change_text = Widgets.TextEntrySet("Change text")
        change_text.add_callback('activated', lambda r: w.set_text(r.get_text()))
        vbox.add_widget(change_text)
        vbox.add_widget(button)

    elif wname == 'expander':
        w = Widgets.Expander(title="Expander Title")
        layout = Widgets.VBox()
        layout.add_widget(Widgets.Label("Expander content"))
        layout.add_widget(Widgets.Button("Button"))
        w.set_widget(layout)
        # Testing set_widget
        button = Widgets.Button("Click to test set_widget")
        button.add_callback('activated', lambda r: w.set_widget(Widgets.Label("Set widget")))
        vbox.add_widget(button)
        vbox.add_widget(w)

    elif wname == 'hbox':
        w = Widgets.HBox()
        w.add_widget(Widgets.Label("Item 1"), stretch=1)
        w.add_widget(Widgets.Label("Item 2"), stretch=1)
        # Test add_widget
        grid = Widgets.GridBox(rows=2, columns=1)
        button = Widgets.Button("Click to add widget")
        button.add_callback('activated', lambda r: w.add_widget(Widgets.Label("Item 3"), stretch=1))
        grid.add_widget(button, 0, 0)
        # Test insert_widget
        label = Widgets.Label("Enter index to insert a widget: ")
        insert = Widgets.TextEntrySet()
        insert.add_callback('activated', lambda r: w.insert_widget(int(r.get_text()), Widgets.Label("Inserted widget"), stretch=1))
        hbox = Widgets.HBox()
        hbox.add_widget(label)
        hbox.add_widget(insert)
        grid.add_widget(hbox, 1, 0)
        vbox.add_widget(w)
        vbox.add_widget(grid)

    elif wname == 'vbox':
        w = Widgets.VBox()
        w.add_widget(Widgets.Label("Item 1"), stretch=0)
        w.add_widget(Widgets.Label("Item 2"), stretch=1)
        # Test add_widget
        grid = Widgets.GridBox(rows=2, columns=1)
        button = Widgets.Button("Click to add widget")
        button.add_callback('activated', lambda r: w.add_widget(Widgets.Label("Item 3"), stretch=1))
        grid.add_widget(button, 0, 0)
        # Test insert_widget
        label = Widgets.Label("Enter index to insert a widget: ")
        insert = Widgets.TextEntrySet()
        insert.add_callback('activated', lambda r: w.insert_widget(int(r.get_text()), Widgets.Label("Inserted widget"), stretch=1))
        hbox = Widgets.HBox()
        hbox.add_widget(label)
        hbox.add_widget(insert)
        grid.add_widget(hbox, 1, 0)
        vbox.add_widget(w)
        vbox.add_widget(grid)

    elif wname == 'splitter':
        w = Widgets.Splitter(orientation='horizontal')
        w.add_widget(Widgets.Label('Content of Pane 1'))
        w.add_widget(Widgets.TextArea())
        # w.add_widget(Widgets.Slider())
        w.add_widget(Widgets.Button("????"))

        w.set_sizes([10, 20, 20])
        vbox.add_widget(w, stretch=1)

    elif wname == 'scrollarea':
        grid = Widgets.GridBox(rows=2, columns=2)
        # Test scroll_to_end
        b = Widgets.Button("Scroll to bottom")
        b.add_callback('activated', lambda r: w.scroll_to_end())
        grid.add_widget(b, 0, 0)
        b = Widgets.Button("Scroll to right")
        b.add_callback('activated', lambda r: w.scroll_to_end(vertical=False, horizontal=True))
        grid.add_widget(b, 0, 1)
        # Test scroll_to_pct
        vert_pct = Widgets.TextEntrySet()
        vert_pct.add_callback('activated', lambda r: w.scroll_to_pct(int(r.get_text())))
        grid.add_widget(vert_pct, 1, 0)
        hori_pct = Widgets.TextEntrySet()
        hori_pct.add_callback('activated', lambda r: w.scroll_to_pct(int(r.get_text()), vertical=False, horizontal=True))
        grid.add_widget(hori_pct, 1, 1)
        vbox.add_widget(grid)
        # Adding multiple widgets to test scrollarea
        a = Widgets.Label("Hello!")
        s = Widgets.Slider(orientation='horizontal')
        s.set_limits(-10, 10, incr_value=1)
        s.set_value(4)
        s.set_tracking(True)
        w = Widgets.ScrollArea()
        img = Widgets.Image()
        img.load_file(os.path.join(icondir, 'ginga-512x512.png'))
        _vbox = Widgets.VBox()
        _vbox.add_widget(img, stretch=0)
        _vbox.add_widget(a, stretch=0)
        _vbox.add_widget(s, stretch=0)
        w.set_widget(_vbox)
        vbox.add_widget(w, stretch=1)

    elif wname == 'tabwidget':
        w = Widgets.TabWidget()
        w.add_widget(Widgets.Label('Content of Tab 1'), title='Tab 1')
        w.add_widget(Widgets.Label('Content of Tab 2'), title='Tab 2')
        hbox = Widgets.HBox()
        sbox = Widgets.SpinBox(dtype=int)
        sbox.set_limits(0, 1, incr_value=1)
        sbox.set_value(0)
        sbox.add_callback('value-changed', lambda sbx, val: w.set_index(val))
        hbox.add_widget(sbox)
        vbox.add_widget(w, stretch=1)
        vbox.add_widget(hbox, stretch=0)

    elif wname == 'stackwidget':
        w = Widgets.StackWidget()
        w.add_widget(Widgets.TextArea(editable=True))
        w.add_widget(Widgets.Label('Content of Stack 1'))
        # Testing set_index
        hbox = Widgets.HBox()
        label = Widgets.Label("Set index to be visible: ")
        set_index = Widgets.TextEntrySet()
        set_index.add_callback('activated', lambda r: w.set_index(int(r.get_text())))
        hbox.add_widget(label)
        hbox.add_widget(set_index)
        # Testing add_widget
        button2 = Widgets.Button("Press to add another widget")
        button2.add_callback('activated', lambda r: w.add_widget(Widgets.Label("Content of inserted widget")))

        vbox.add_widget(hbox, stretch=0)
        vbox.add_widget(button2, stretch=0)
        vbox.add_widget(w, stretch=1)

    elif wname == 'mdiwidget':
        w = Widgets.MDIWidget()
        w.add_widget(Widgets.Label('Content of MDI Area 1'))
        w.add_widget(Widgets.Label('Content of MDI Area 2'))
        vbox.add_widget(w, stretch=1)

    elif wname == 'gridbox':
        w = Widgets.GridBox(rows=2, columns=2)
        w.add_widget(Widgets.Label('Content of Grid Area 1'), 0, 0)
        w.add_widget(Widgets.Label('Content of Grid Area 2'), 0, 1)
        w.add_widget(Widgets.Label('Content of Grid Area 3'), 1, 0)
        w.add_widget(Widgets.Label('Content of Grid Area 4'), 1, 1)
        # Test set row and column spacing methods
        row_button = Widgets.Button("Press to change row spacing to 50px")
        row_button.add_callback('activated', lambda r: w.set_row_spacing(50))
        col_button = Widgets.Button("Press to change column spacing to 50px")
        col_button.add_callback('activated', lambda r: w.set_column_spacing(50))
        both_button = Widgets.Button("Press to change row and column spacing to 20px")
        both_button.add_callback('activated', lambda r: w.set_spacing(20))
        # Test insert_row
        hbox = Widgets.HBox()
        label = Widgets.Label("Enter index to insert a row: ")
        insert_row = Widgets.TextEntrySet()
        insert_row.add_callback('activated', lambda r: w.insert_row(int(r.get_text())))
        hbox.add_widget(label)
        hbox.add_widget(insert_row)
        # Test delete_row
        hbox2 = Widgets.HBox()
        label = Widgets.Label("Enter index to delete a row: ")
        delete_row = Widgets.TextEntrySet()
        delete_row.add_callback('activated', lambda r: w.delete_row(int(r.get_text())))
        hbox2.add_widget(label)
        hbox2.add_widget(delete_row)
        # Test append_row
        append_button = Widgets.Button("Press to append a row")
        append_button.add_callback('activated', lambda r: w.append_row())
        # Test add_widget
        hbox3 = Widgets.HBox()
        label = Widgets.Label("Enter indices to add widget (row,column): ")
        add_widget = Widgets.TextEntrySet()
        add_widget.add_callback('activated', lambda r: grid_helper(r, w))
        hbox3.add_widget(label)
        hbox3.add_widget(add_widget)

        vbox.add_widget(w, stretch=1)
        vbox.add_widget(row_button, stretch=1)
        vbox.add_widget(col_button, stretch=1)
        vbox.add_widget(both_button, stretch=1)
        vbox.add_widget(hbox, stretch=1)
        vbox.add_widget(hbox2, stretch=1)
        vbox.add_widget(hbox3, stretch=1)
        vbox.add_widget(append_button, stretch=1)

    elif wname == 'menubar':
        w = Widgets.Menubar()
        menu = w.add_name('Menu 1')
        menu.add_name('Larry').add_callback('activated',
                                            lambda *args: print("chose Larry"))
        menu.add_name('Curly').add_callback('activated',
                                            lambda *args: logger.info("chose Curly"))
        menu.add_name('Moe').add_callback('activated',
                                          lambda *args: logger.info("chose Moe"))

        menu = w.add_name('3Amigos')
        menu.add_name('Hector').add_callback('activated',
                                             lambda *args: print("chose Hector"))
        menu.add_name('Manuel').add_callback('activated',
                                             lambda *args: logger.info("chose Manuel"))
        menu.add_name('Roberto').add_callback('activated',
                                              lambda *args: logger.info("chose Roberto"))
        vbox.add_widget(w)
        vbox.add_widget(Widgets.Label("App content"), stretch=1)

    elif wname == 'toolbar':
        w = Widgets.Toolbar()
        menu = w.add_menu('Menu Type 1', mtype='tool')
        menu.add_name('Larry').add_callback('activated',
                                            lambda w: logger.info("chose Larry"))
        menu.add_name('Curly').add_callback('activated',
                                            lambda w: logger.info("chose Curly"))
        menu.add_name('Moe').add_callback('activated',
                                          lambda w: logger.info("chose Moe"))
        menu = w.add_menu('Menu Type 2', mtype='mbar')
        menu.add_name('Frank')
        menu.add_name('Dean')
        menu.add_name('Sammy')
        w.add_widget(Widgets.Button('A Button'))
        w.add_separator()
        w.add_action("Toggle me", toggle=True)
        w.add_action(None, iconpath=os.path.join(icondir, 'hand.svg'))
        vbox.add_widget(w)
        vbox.add_widget(Widgets.Label("App content"), stretch=1)

    elif wname == 'dialog':
        dia = Widgets.Dialog(title="Dialog Title",
                             buttons=[('ok', 0), ('cancel', 1)],
                             parent=top, modal=False)
        dia.add_callback('activated',
                         lambda w, rsp: logger.info("user chose %s" % (rsp)))
        top.add_dialog(dia)
        cntr = dia.get_content_area()
        cntr.add_widget(Widgets.Label("My Dialog Content"))

        # add some content to main app widget
        w = Widgets.Label("Hello World label")
        vbox.add_widget(w, stretch=1)
        hbox = Widgets.HBox()
        w = Widgets.Button("Open Dialog")
        w.add_callback('activated', lambda w: dia.show())
        hbox.add_widget(w)
        w = Widgets.Button("Close Dialog")
        w.add_callback('activated', lambda w: dia.hide())
        hbox.add_widget(w)
        vbox.add_widget(hbox)

    else:
        # default to label
        logger.error("Don't understand kind of widget '%s'" % (wname))
        w = Widgets.Label("Hello World label")
        vbox.add_widget(w, stretch=1)

    dia = Widgets.Dialog(title="Example: " + wname,
                         parent=top, modal=False)
    dia.add_callback('close', lambda *args: dia.delete())
    cntr = dia.get_content_area()
    cntr.add_widget(vbox)
    if hasattr(top, 'add_dialog'):
        top.add_dialog(dia)
    dia.show()


def main(options, args):
    logger = log.get_logger('test', options=options)

    import ginga.toolkit as ginga_toolkit

    ginga_toolkit.use(options.toolkit)

    from ginga.gw import Widgets

    #app = Widgets.Application(host='', logger=logger)
    app = Widgets.Application(logger=logger)
    app.add_callback('shutdown', quit)

    page = None
    if hasattr(app, 'script_imports'):
        app.script_imports.append('jqx')

        top = Widgets.Page(title="Ginga Wrapped Widgets")
        app.add_window(top)
    else:
        top = app.make_window("Ginga Wrapped Widgets")

    top.add_callback('close', quit)

    hbox = Widgets.HBox()
    hbox.set_border_width(2)
    hbox.set_spacing(1)

    cbox = Widgets.ComboBox()
    for wname in ['label', 'button', 'textentry', 'textentryset', 'textarea',
                  'checkbox', 'togglebutton', 'radiobutton', 'combobox',
                  'spinbox', 'slider', 'scrollbar', 'progressbar', 'statusbar',
                  'image', 'treeview', 'webview', 'frame', 'expander',
                  'hbox', 'vbox', 'splitter', 'scrollarea', 'tabwidget',
                  'stackwidget', 'mdiwidget', 'gridbox', 'menubar', 'toolbar',
                  'dialog', 'dial']:
        cbox.insert_alpha(wname)

    hbox.add_widget(cbox)
    btn = Widgets.Button("Show Example")
    btn.add_callback('activated', lambda w: show_example(cbox, top, logger))
    hbox.add_widget(btn)

    top.set_widget(hbox)
    top.show()
    top.raise_()

    try:
        app.mainloop()

    except KeyboardInterrupt:
        print("Terminating viewer...")
        if top is not None:
            top.close()


if __name__ == "__main__":
    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt|pg)")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    main(options, args)



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os
import glob
import inspect

from ginga.util import paths, compat
from ginga.modes.mode_base import Mode
from ginga.misc.ModuleManager import my_import

available_modes = []


def add_mode(mode_class):
    if mode_class not in available_modes:
        available_modes.append(mode_class)


def discover_modes():
    # discover modes installed under entry point "ginga_modes"
    group = 'ginga_modes'
    discovered_modes = compat.ep_get(group)
    for entry_point in discovered_modes:
        try:
            mode_class = entry_point.load()
            add_mode(mode_class)

        except Exception as e:
            print("Error trying to load entry point %s: %s" % (
                str(entry_point), str(e)))

    # discover modes in the user's $HOME/.ginga/modes folder
    moduledir = os.path.join(paths.ginga_home, "modes")
    if os.path.isdir(moduledir):
        files = glob.glob(os.path.join(moduledir, "*.py"))
        for path in files:
            _dir, fname = os.path.split(path)
            fname, _sfx = os.path.splitext(fname)
            module = my_import(fname, path=path)
            for attr in module.__dict__.values():
                if (inspect.isclass(attr) and issubclass(attr, Mode) and
                    attr is not Mode):
                    try:
                        add_mode(attr)

                    except Exception as e:
                        print("Error trying to load mode %s: %s" % (
                            fname, str(e)))


if len(available_modes) == 0:
    discover_modes()
